import re

import logging
log = logging.getLogger(__name__)

import coralme
from collections import Counter

def get_remaining_complex_elements(model, met, modification_formulas):
	# get base complex and modifications, without compartment ID
	regex = '_mod_([A-Za-z0-9_]*\(\d*\))'
	regex = re.compile(regex)

	tmp_met = coralme.core.component.Metabolite('tmp_met')
	base_complex = met.id.split('_mod_')[0]
	components = re.findall(regex, met.id)
	elements = Counter()

	# If a the completely unmodified complex is present in the model
	# has a formula, initialize the elements dictionary with that
	mets = model.metabolites
	if base_complex in mets and mets.get_by_id(base_complex).formula:
		elements.update(mets.get_by_id(base_complex).elements)

	# get component and stoichiometry per modification
	regex = '([A-Za-z0-9_]*)\((\d*)\)'
	regex = re.compile(regex)
	for component in components:
		new_elements = elements.copy()
		new_complex = '_mod_'.join([base_complex, component])

		if new_complex in mets and mets.get_by_id(new_complex).formula:
			# default to new_complex elements if both new and old exist
			if base_complex in mets and mets.get_by_id(base_complex).formula:
				new_elements = Counter()
			formula = mets.get_by_id(new_complex).formula
			tmp_met.formula = formula
			new_elements.update(tmp_met.elements)

		## Net effect of an SH modification is adding a Sulfur to elements
		## SH should be in the me_mets file
		#elif 'SH(1)' in component:
			#new_elements['S'] += 1

		elif 'glycyl(1)' in component:
			new_elements['H'] -= 1

		# modifies O- to SH
		elif component == 'cosh(1)':
			new_elements['O'] -= 1
			new_elements['S'] += 1
			new_elements['H'] += 1

		elif component in modification_formulas: # Previously, a cofactor stoichiometry of 1 could be omitted
			formula = modification_formulas[component]
			tmp_met.formula = formula
			new_elements.update(tmp_met.elements)

		elif 'Oxidized(1)' in component and 'FLAVODOXIN' not in base_complex:
			new_elements.update({'H': -2})

		elif '(' in component: # FIXED: Check new complex ID convention
			component, value = re.findall(regex, component)[0]
			if component in modification_formulas:
				tmp_met.formula = modification_formulas[component]
			elif component + '_c' in mets:
				tmp_met.formula = mets.get_by_id(component + '_c').formula
			else:
				logging.warning('No formula found for modification \'{:s}\' either in me.metabolites nor metabolites.txt input.'.format(component))
				continue

			for e, v in tmp_met.elements.items():
				new_elements[e] += v * float(value)

		if elements == new_elements and 'FLAVODOXIN' not in base_complex:
			logging.warning('The stoichiometry of \'{:s}\' is identical to \'{:s}\'. Check if it is the correct behavior.'.format(met.id, base_complex))

		base_complex = '_mod_'.join([base_complex, component]) # DO NOT DELETE
		elements = new_elements.copy()

	return elements

def add_remaining_complex_formulas(model, modification_formulas):
	"""
	Add formula to complexes that are not formed from a complex formation
	reaction (i.e., complexes involved in metabolic reactions)
	"""
	element_dict = {}

	# Reset all formulas first
	complex_list = []
	for met in model.metabolites:
		# If met is not a complex or not formed by a complex formation reaction, do not reset
		if not isinstance(met, coralme.core.component.Complex) or met.id in model.process_data:
			continue

		for r in met.reactions:
			if hasattr(r, 'update'):
				r.update()

		met.formula = None
		met.elements = {}
		complex_list.append(met)

	# Get formulas only for complexes without complex formation reaction
	for met in complex_list:
		element_dict[met] = get_remaining_complex_elements(model, met, modification_formulas)

	# Adding elements for complexes dynamically can change function output
	# Update all of them after
	for met, elements in element_dict.items():
		coralme.util.massbalance.elements_to_formula(met, elements)



import logging
log = logging.getLogger(__name__)
import coralme

def add_iron_sulfur_modifications(me_model):
	#generic_fes_transfer_complexes = me_model.global_info['complex_cofactors']['generic_fes_transfer_complexes']
	fes_transfers = me_model.global_info['complex_cofactors']['fes_transfers']

	#identify if the user added MetabolicReactions to create modified FeS transfers
	#if not, FormationReaction's are modified during update(),
	#but missing 2fe2s and 4fe4s will be added as troubleshooted reactions
	for fes in ['2fe2s', '4fe4s']:
		# WARNING: If the organism doesn't have iron-sulfur loaders, CPLX_dummy will be added as an iron-sulfur loader
		# WARNING: The troubleshooter will add sink reactions for 4fe4s/2fe2s, making necessary formation reactions
		# See main.py, lines 1931-1936
		components = [ '{:s}_mod_{:s}(1)'.format(x, fes) for x in set(fes_transfers) if x != '' and x != 'CPLX_dummy' ]
		for component in components:
			query = me_model.metabolites.get_by_id(component).reactions
			query = [ x for x in query if x.metabolites[me_model.metabolites.get_by_id(component)] > 0 ]
			rtypes = [ type(x) for x in query ]
			# two type of reactions if user added the correct formation reaction in reaction_matrix.txt
			if coralme.core.reaction.MetabolicReaction in rtypes and coralme.core.reaction.ComplexFormation in rtypes:
				# we need to remove subreactions from process_data
				to_modify = [ x for x in query if isinstance(x, coralme.core.reaction.ComplexFormation)][0]
				me_model.process_data.get_by_id(to_modify.id.replace('formation_', '')).subreactions.pop('mod_{:s}_c'.format(fes))
				# and remove the formation reaction of the modified complex only
				me_model.process_data.remove(to_modify.id.replace('formation_', ''))
				me_model.remove_reactions([to_modify])

	for fes in ['2fe2s', '4fe4s']:
		name = 'generic_{:s}_transfer_complex'.format(fes)
		components = [ '{:s}_mod_{:s}(1)'.format(x, fes) for x in set(fes_transfers) if x != '' ]
		for component in components:
			if not me_model.metabolites.has_id(component):
				me_model.add_metabolites([coralme.core.component.Complex(component)])

		generic_fes_transfer = coralme.core.processdata.GenericData(name, me_model, components)
		generic_fes_transfer.create_reactions()

		# add 2fe2s or 4fe4s to the model if it doesn't exist
		me_model.add_metabolites([coralme.core.component.Metabolite(fes + '_c')])

		# create unloading reactions
		for name in set(fes_transfers):
			if name != '':
				rxn = coralme.core.reaction.MEReaction('_'.join([name, fes, 'unloading']))
				me_model.add_reactions([rxn])
				rxn.add_metabolites({ name + '_mod_' + fes + '(1)': -1, fes + '_c': 1, name: 1 })

	# add fes transfer enzymes to proper modification data
	if me_model.process_data.has_id('mod_2fe2s_c'):
		mod_2fe2s = me_model.process_data.mod_2fe2s_c
		mod_2fe2s.enzyme = ['generic_2fe2s_transfer_complex']
		mod_2fe2s.stoichiometry = { '2fe2s_c': -1 }
		mod_2fe2s._element_contribution = { 'Fe': 2, 'S': 2 }

	if me_model.process_data.has_id('mod_4fe4s_c'):
		mod_4fe4s = me_model.process_data.mod_4fe4s_c
		mod_4fe4s.enzyme = ['generic_4fe4s_transfer_complex']
		mod_4fe4s.stoichiometry = { '4fe4s_c': -1 }
		mod_4fe4s._element_contribution = { 'Fe': 4, 'S': 4 }

	if me_model.process_data.has_id('mod_3fe4s_c'):
		mod_3fe4s = me_model.process_data.mod_3fe4s_c
		mod_3fe4s.enzyme = ['generic_4fe4s_transfer_complex']
		mod_3fe4s.stoichiometry = { '4fe4s_c': -1, 'fe2_c': 1 }
		mod_3fe4s._element_contribution = { 'Fe': 3, 'S': 4 }

	fes_chaperones = me_model.global_info['complex_cofactors'].get('fes_chaperones', {})
	for chaperone in set(fes_chaperones.values()):
		new_mod = coralme.core.processdata.SubreactionData('mod_2fe2s_c_' + chaperone, me_model)
		new_mod.enzyme = [ chaperone, 'generic_2fe2s_transfer_complex' ]
		new_mod.stoichiometry = { '2fe2s_c': -1 }

	if me_model.process_data.has_id('mod_2fe2s_c'):
		for cplx_data in me_model.process_data.get_by_id('mod_2fe2s_c').get_complex_data():
			cplx_id = cplx_data.id.split('_mod')[0]
			if cplx_id in fes_chaperones:
				cplx_data.subreactions[ 'mod_2fe2s_c_' + fes_chaperones[cplx_id] ] = cplx_data.subreactions.pop('mod_2fe2s_c')

def _replace_modification(dct, me_model):
	modification = list(dct.keys())[0] # current modification ID in me.process_data
	new_subrxn_id = list(dct.values())[0][0] # subreaction ID in process_data to replace modification ID

	if me_model.process_data.has_id(new_subrxn_id):
		new_mod_data = me_model.process_data.get_by_id(new_subrxn_id)
	else: # create SubreactionData with new ID (e.g., mod_btn_c now biotin_ligase)
		new_mod_data = coralme.core.processdata.SubreactionData(new_subrxn_id, me_model)
		try:
			new_mod_data._element_contribution = new_mod_data.calculate_element_contribution()
		except:
			logging.warning('All metabolites in SubreactionData \'{:s}\' must have a formula to determine their elemental contribution.'.format(new_mod_data))

	# This function does not replace information in process_data; it rather changes information in subreactions
	for data in me_model.process_data.get_by_id(modification).get_complex_data():
		cplx_data = me_model.process_data.get_by_id(data.id)
		# copy the original data
		cplx_data.complex_id = data.complex_id
		cplx_data.stoichiometry = data.stoichiometry
		cplx_data.subreactions = data.subreactions.copy()
		# e.g.: remove mod_btn_c and replace it with the new subreaction id
		cplx_data.subreactions[new_mod_data.id] = cplx_data.subreactions.pop(modification)
		#cplx_data.create_complex_formation() # the cplx_data already exists

def add_btn_modifications(me_model):
	# dct = { "mod_btn_c" : [ "biotin_ligase" ] }
	dct = me_model.global_info['complex_cofactors'].get('biotin_subreactions', {})
	if bool(dct):
		_replace_modification(dct, me_model)

def add_2tpr3dpcoa_modifications(me_model):
	# dct = { "mod_2tpr3dpcoa_c" : [ "citx_transfer_to_citd" ] }
	dct = me_model.global_info['complex_cofactors'].get('citx_subreactions', {})
	if bool(dct):
		_replace_modification(dct, me_model)

def add_glycyl_modifications(me_model):
	# dct = { "mod_glycyl_c" : [ "gre_activation" ] }
	dct = me_model.global_info['complex_cofactors'].get('glycyl_subreactions', {})

	if dct:
		# WARNING: Do we need to correct all the modification.enzyme properties (biotin, 2tpr3dpcoa, pan4p)?
		data = me_model.subreaction_data.get_by_id(list(dct.values())[0][0])
		for met, stoich in data.stoichiometry.items():
			if isinstance(me_model.metabolites.get_by_id(met), coralme.core.component.Complex) and stoich < 0:
				data.enzyme.add(met)

	if bool(dct):
		_replace_modification(dct, me_model)

def add_pan4p_modifications(me_model):
	# dct = { "mod_pan4p_c" : [ "acpP_activation" ] }
	dct = me_model.global_info['complex_cofactors'].get('acps_subreactions', {})
	if bool(dct):
		_replace_modification(dct, me_model)

def add_FeFe_and_NiFe_modifications(me_model):
	# dct = { "mod_FeFe_cofactor_c" : [ "base_complex" ], "mod_NiFe_cofactor_c" : [ "base_complex" ] }
	dct = me_model.global_info['complex_cofactors'].get('FeFe/NiFe', {})

	for mod, base_complexes in dct.items():
		for base_complex in base_complexes:
			if base_complex != '' and me_model.process_data.has_id(mod):
				complex_data = list(me_model.process_data.get_by_id(mod).get_complex_data())
				if len(complex_data) > 0:
					for data in complex_data:
						cplx_data = me_model.process_data.get_by_id(data.id)
						cplx_data.complex_id = data.complex_id
						cplx_data.stoichiometry = { base_complex + '_' + mod[:-2] + '(1)' : 1 }
						# The modifications occurs now as follow:
						# base_complex + FeFe/NiFe => base_complex_mod_FeFe/NiFe
						# base_complex_mod_FeFe/NiFe + other cofactors => final modified complex
						cplx_data.subreactions[mod] = 0
				else:
					logging.warning('The ID \'{:s}\' in the configuration file has no base complexes assigned to it.'.format(mod))
			else:
				logging.warning('The ID \'{:s}\' in the configuration file does not exist in the ME-model.'.format(mod))

#def add_lipoate_modifications(me_model, lipoate_modifications):
def add_lipoyl_modifications(me_model):
	# Two different reactions can add a lipoate modification.
	# We create a separate SubreactionData for each one
	#for key, mod in lipoate_modifications):
	lipoate_modifications = me_model.global_info['complex_cofactors']['lipoate_subreactions']
	for mod in list(lipoate_modifications.values())[0]:
		if mod in me_model.process_data:
			mod_data = me_model.process_data.get_by_id(mod)
		else:
			# create SubreactionData
			mod_data = coralme.core.processdata.SubreactionData(mod, me_model)

		#info = me_model.process_data.get_by_id(mod)
		#mod_data.enzyme = mod_data.enzyme
		#mod_data.stoichiometry = mod_data.stoichiometry
		# element count for lipoate modifications
		try:
			mod_data._element_contribution = mod_data.calculate_element_contribution()
		except:
			logging.warning('All metabolites in SubreactionData \'{:s}\' must have a formula to determine their elemental contribution.'.format(mod))

	lipoate = me_model.process_data.get_by_id(list(lipoate_modifications.keys())[0])
	#alt_lipo = me_model.process_data.get_by_id('mod_lipo_c_alt')
	for data in lipoate.get_complex_data():
		#alt_cplx_data = coralme.core.processdata.ComplexData(data.id + '_alt', me_model)
		#alt_cplx_data.complex_id = data.complex_id
		#alt_cplx_data.stoichiometry = data.stoichiometry
		#alt_cplx_data.subreactions = data.subreactions.copy()
		#alt_cplx_data.subreactions[alt_lipo.id] = alt_cplx_data.subreactions.pop(lipoate.id)
		#alt_cplx_data.create_complex_formation()

		# We need to replace the `mod_lipoyl_c` (the key in lipoate_modifications) by the values in lipoate_modifications
		for mod in list(lipoate_modifications.values())[0]:
			new_cplx_data = coralme.core.processdata.ComplexData(data.id + '_' + mod, me_model)
			# copy the original data
			new_cplx_data.complex_id = data.complex_id
			new_cplx_data.stoichiometry = data.stoichiometry
			new_cplx_data.subreactions = data.subreactions.copy()
			# remove mod_lipoyl_c and replace it with the new subreaction id
			new_cplx_data.subreactions[mod] = new_cplx_data.subreactions.pop(lipoate.id)
			new_cplx_data.create_complex_formation()

	# Delete old ComplexFormation data
	# TODO: remove entry from process_data
	lst = list(me_model.process_data.get_by_id('mod_lipoyl_c').get_complex_data())
	lst = [ x.formation for x in lst ]
	me_model.remove_reactions(lst)
	#me_model.process_data.remove('mod_lipoyl_c')

def add_bmocogdp_chaperones(me_model):
	bmocogdp_chaperones = me_model.global_info['complex_cofactors']['bmocogdp_chaperones']
	for chaperone in set(bmocogdp_chaperones.values()):
		new_mod = coralme.core.processdata.SubreactionData('mod_bmocogdp_c_' + chaperone, me_model)
		new_mod.enzyme = [chaperone]
		new_mod.stoichiometry = {'bmocogdp_c': -1}

	for cplx_data in me_model.process_data.get_by_id('mod_bmocogdp_c').get_complex_data():
		cplx_id = cplx_data.id.split('_mod')[0]
		if cplx_id in bmocogdp_chaperones:
			cplx_data.subreactions['mod_bmocogdp_c_' + bmocogdp_chaperones[cplx_id]] = cplx_data.subreactions.pop('mod_bmocogdp_c')

# OLD CODE not used anymore
def add_modification_procedures(me_model):
	# add SubreactionData for iron sulfur clusters
	add_iron_sulfur_modifications(me_model)

	# lipoate modifications can be accomplished using two different mechanisms
	add_lipoate_modifications(me_model)

	# bmocogdp modifications have multiple selective chaperones that transfer
	# the metabolite to the target complexes
	add_bmocogdp_modifications(me_model)




#!/usr/bin/python3

import collections
import json
import os

def dict_to_defaultdict(dct):
	"""Cast a dict to a defaultdict"""
	return collections.defaultdict(lambda: [], dct)

def save_curation_notes(curation_notes,filepath):
	"""Save curation notes as JSON"""
	file = open(filepath,'w')
	file.write(json.dumps(curation_notes, indent=4))
	file.close()

def load_curation_notes(filepath):
	"""Load curation notes from a JSON"""
	if not os.path.isfile(filepath):
		return collections.defaultdict(list)
	with open(filepath) as json_file:
		return json.load(json_file,object_hook=dict_to_defaultdict)

def publish_curation_notes(curation_notes,filepath):
	"""Print the curation notes to a txt file"""
	file = open(filepath,'w')
	for k,v in curation_notes.items():
		file.write('\n')
		for w in v:
			file.write('{} {}@{} {}\n'.format('#'*20,w['importance'],k,'#'*20))
			file.write('{} {}\n'.format('*'*10,w['msg']))
			if 'triggered_by' in w:
				file.write('The following items triggered the warning:\n')
				for i in w['triggered_by']:
					if isinstance(i,dict):
						file.write('\n')
						file.write(json.dumps(i))
						file.write('\n')
					else:
						file.write(i + '\n')
			file.write('\n{}Solution:\n{}\n\n'.format('*'*10,w['to_do']))
		file.write('\n\n')
	file.close()



import tqdm
bar_format = '{desc:<75}: {percentage:.1f}%|{bar}| {n_fmt:>5}/{total_fmt:>5} [{elapsed}<{remaining}]'

import sympy
import coralme

import logging
log = logging.getLogger(__name__)

from warnings import warn

_REQUIRED_REACTION_ATTRIBUTES = {
	"id",
	"name",
	"metabolites",
	"lower_bound",
	"upper_bound",
	"objective_coefficient",
	"subsystem"
	#"variable_kind"
	}

# Reaction types can have different attributes
_REACTION_TYPE_DEPENDENCIES = {
	'MetabolicReaction': [
		'complex_data',
		'stoichiometric_data',
		'reverse',
		'keff'
		],
	'ComplexFormation': [
		'_complex_id',
		'complex_data_id'
		],
	'PostTranslationReaction': [
		'posttranslation_data'
		],
	'TranscriptionReaction': [
		'transcription_data'
		],
	'GenericFormationReaction': [],
	'MEReaction': [],
	'SummaryVariable': [],
	'TranslationReaction': [
		'translation_data'
		],
	'tRNAChargingReaction': [
		'tRNA_data'
		]
	}

_REQUIRED_PROCESS_DATA_ATTRIBUTES = {
	"id"
	}

# Process data types have different attributes
_PROCESS_DATA_TYPE_DEPENDENCIES = {
	'StoichiometricData': [
		'_stoichiometry',
		'lower_bound',
		'upper_bound',
		'subreactions'
		],
	'ComplexData': [
		'stoichiometry',
		'complex_id',
		'subreactions'],
	'TranscriptionData': [
		'subreactions',
		'nucleotide_sequence',
		'RNA_products',
		'RNA_polymerase',
		'organelle'
		],
	'TranslationData': [
		'subreactions',
		'nucleotide_sequence',
		'mRNA',
		'protein',
		'transl_table',
		'translation',
		'organelle',
		'pseudo'
		],
	'tRNAData': [
		'subreactions',
		'codon',
		'RNA',
		'amino_acid',
		'synthetase',
		'synthetase_keff'
		],
	'TranslocationData': [
		'enzyme_dict',
		'stoichiometry',
		'keff',
		'length_dependent_energy'
		],
	'PostTranslationData': [
		'processed_protein_id',
		'unprocessed_protein_id',
		'propensity_scaling',
		'aggregation_propensity',
		'translocation',
		'subreactions',
		'surface_area',
		'keq_folding',
		'k_folding',
		'biomass_type',
		'translocation_multipliers'
		],
	'SubreactionData': [
		'stoichiometry',
		'enzyme',
		'keff',
		'element_contribution'
		],
	'GenericData': [
		'component_list'
		]
	}

_REQUIRED_METABOLITE_ATTRIBUTES = {
	"id",
	"name",
	"formula",
	"compartment"
	}

_OPTIONAL_METABOLITE_ATTRIBUTES = {
	"charge",
	"_bound",
	"_constraint_sense"
	}

# Some metabolite types require additional attributes
_METABOLITE_TYPE_DEPENDENCIES = {
	'TranscribedGene': [
		'left_pos',
		'right_pos',
		'strand',
		'RNA_type',
		'nucleotide_sequence'
		],
	'ProcessedProtein': [
		'unprocessed_protein_id'
		]
	}

def get_sympy_expression(value, growth_key):
	"""
	Return sympy expression from json string using sympify

	mu is assumed to be positive but using sympify does not apply this
	assumption. The mu symbol produced from sympify is replaced with
	coralme's mu value to ensure the expression can be used in the model.

	Parameters
	----------
	value : str
		String representation of mu containing expression

	Returns
	-------
	sympy expression
		Numeric representation of string with coralme's mu symbol substituted

	"""
	# The json file includes the 'mu' key in dct['global_info']['growth_key'] as a string
	# We use dct['global_info']['growth_key'] to set a sympy.Symbol called 'growth_key'
	expression_value = sympy.sympify(value)
	if isinstance(expression_value, (sympy.core.numbers.Float, sympy.core.numbers.One, sympy.core.numbers.NegativeOne, sympy.core.numbers.Integer)):
		return float(expression_value)
	else:
		return expression_value.subs(str(growth_key), growth_key)

def get_numeric_from_string(string):
	"""

	Parameters
	----------
	string : str
		String representation of numeric expression

	Returns
	-------
	float or sympy expression
		Numeric representation of string

	"""
	try:
		return float(string)
	except ValueError:
		return get_sympy_expression(string)

def _fix_type(value):
	"""convert possible types to str, float, and bool"""
	# Because numpy floats can not be pickled to json
	if isinstance(value, str):
		return str(value)
	if isinstance(value, float):
		return float(value)
	if isinstance(value, bool):
		return bool(value)
	if isinstance(value, set):
		return list(value)
	if isinstance(value, sympy.Basic):
		return str(value)
	if hasattr(value, 'id'):
		return str(value.id)
	# if value is None:
	#	 return ''
	return value

def _reaction_to_dict(reaction):
	new_reaction = { key:_fix_type(getattr(reaction, key)) for key in _REQUIRED_REACTION_ATTRIBUTES if key != 'metabolites' }

	reaction_type = reaction.__class__.__name__
	new_reaction['reaction_type'] = {}
	new_reaction['reaction_type'][reaction_type] = {}

	for attribute in _REACTION_TYPE_DEPENDENCIES.get(reaction_type, []):
		reaction_attribute = getattr(reaction, attribute)
		new_reaction['reaction_type'][reaction_type][attribute] = _fix_type(reaction_attribute)

	# Add metabolites
	new_reaction['metabolites'] = {}
	for met, value in reaction.metabolites.items():
		new_reaction['metabolites'][met.id] = _fix_type(value)

	return new_reaction

def _process_data_to_dict(data):
	process_data_type = data.__class__.__name__

	new_data = {
		key: _fix_type(getattr(data, key))
		for key in _REQUIRED_PROCESS_DATA_ATTRIBUTES
		}

	new_data['process_data_type'] = {}
	new_data['process_data_type'][process_data_type] = {}
	new_process_data_type_dict = new_data['process_data_type'][process_data_type]

	special_list = [
		'subreactions',
		'stoichiometry',
		'enzyme_dict',
		'surface_area',
		'keq_folding',
		'k_folding'
		]

	for attribute in _PROCESS_DATA_TYPE_DEPENDENCIES[process_data_type]:
		if attribute not in special_list:
			data_attribute = getattr(data, attribute)
			new_process_data_type_dict[attribute] = _fix_type(data_attribute)

		elif attribute == 'enzyme_dict':
			new_process_data_type_dict[attribute] = {}
			for cplx, values in getattr(data, attribute).items():
				new_process_data_type_dict[attribute][cplx] = {}
				for property, value in values.items():
					new_process_data_type_dict[attribute][cplx][property] = _fix_type(value)
		else:
			new_process_data_type_dict[attribute] = {}
			for metabolite, coefficient in getattr(data, attribute).items():
				new_process_data_type_dict[attribute][metabolite] = _fix_type(coefficient)

	return new_data

def _metabolite_to_dict(metabolite):
	metabolite_type = metabolite.__class__.__name__
	new_metabolite = {
		key: _fix_type(getattr(metabolite, key))
		for key in _REQUIRED_METABOLITE_ATTRIBUTES
		}

	# Some metabolites require additional information to construct working
	# ME-model
	new_metabolite['metabolite_type'] = {}
	new_metabolite['metabolite_type'][metabolite_type] = {}
	for attribute in _METABOLITE_TYPE_DEPENDENCIES.get(metabolite_type, []):
		metabolite_attribute = getattr(metabolite, attribute)
		new_metabolite['metabolite_type'][metabolite_type][attribute] = metabolite_attribute

	return new_metabolite

def _get_attribute_array(dictlist, type):
	if type == 'reaction':
		return [_reaction_to_dict(reaction) for reaction in dictlist]
	elif type == 'process_data':
		return [_process_data_to_dict(data) for data in dictlist]
	elif type == 'metabolite':
		return [_metabolite_to_dict(metabolite) for metabolite in dictlist]
	else:
		raise TypeError('Type must be reaction, process_data or metabolite')

def _get_global_info_dict(global_info):
	new_global_info = {}
	for key, value in global_info.items():
		if type(value) != dict:
			new_global_info[key] = _fix_type(value)
		else:
			new_global_info[key] = value
	return new_global_info

def me_model_to_dict(model):
	"""
	Create dictionary representation of full ME-model

	Parameters
	----------
	model : :class:`~coralme.core.model.MEModel`

	Returns
	-------
	dict
		Dictionary representation of ME-model

	"""

	obj = dict(
		reactions=_get_attribute_array(model.reactions, 'reaction'),
		process_data=_get_attribute_array(model.process_data, 'process_data'),
		metabolites=_get_attribute_array(model.metabolites, 'metabolite'),
		global_info=_get_global_info_dict(model.global_info)
	)

	return obj

# -----------------------------------------------------------------------------
# Functions below here are used to create a ME-model from its dictionary
# representation

def _add_metabolite_from_dict(model, metabolite_info):
	"""
	Builds metabolite instances defined in dictionary, then add it to the
	ME-model being constructed.

	ProcessedProteins require additional information
	"""

	metabolite_type_dict = metabolite_info['metabolite_type']
	if len(metabolite_type_dict) != 1:
		raise Exception('Only 1 metabolite_type in valid json')

	metabolite_type = list(metabolite_type_dict.keys())[0]

	# ProcessedProtein types require their unprocessed protein id as well
	if metabolite_type == 'ProcessedProtein':
		unprocessed_id = metabolite_type_dict['ProcessedProtein']['unprocessed_protein_id']

		metabolite_obj = getattr(coralme.core.component, metabolite_type)(metabolite_info['id'], unprocessed_id)

	elif metabolite_type == 'TranscribedGene':
		rna_type = metabolite_type_dict['TranscribedGene']['RNA_type']
		nucleotide_sequence = metabolite_type_dict['TranscribedGene']['nucleotide_sequence']
		metabolite_obj = getattr(coralme.core.component, metabolite_type)(metabolite_info['id'], rna_type, nucleotide_sequence)
	else:
		metabolite_obj = getattr(coralme.core.component, metabolite_type)(metabolite_info['id'])

	for attribute in _REQUIRED_METABOLITE_ATTRIBUTES:
		setattr(metabolite_obj, attribute, metabolite_info[attribute])

	for attribute in _METABOLITE_TYPE_DEPENDENCIES.get(metabolite_type, []):
		value = metabolite_type_dict[metabolite_type][attribute]
		setattr(metabolite_obj, attribute, value)

	model.add_metabolites([metabolite_obj])

def _add_process_data_from_dict(model, process_data_dict):
	"""
	Builds process_data instances defined in dictionary, then add it to the
	ME-model being constructed.

	Most classes of process_data only require an id and model to initiate them,
	but TranslationData, tRNAData, PostTranslationData and GenericData require
	additional inputs.

	"""

	# Create process data instances. Handle certain types individually
	id = process_data_dict['id']
	process_data_type_dict = process_data_dict['process_data_type']
	if len(process_data_type_dict) == 1:
		process_data_type, process_data_info = process_data_type_dict.popitem()
	else:
		print(process_data_type_dict, len(process_data_type_dict))
		raise Exception('Only 1 reaction_type in valid json')

	if process_data_type == 'TranscriptionData':
		nucleotide_sequence = process_data_info['nucleotide_sequence']
		rnap = process_data_info['RNA_polymerase']
		rna_products = process_data_info['RNA_products']
		organelle = process_data_info['organelle']
		process_data = getattr(coralme.core.processdata, process_data_type)(id, model, nucleotide_sequence, rnap, rna_products, organelle)
	elif process_data_type == 'TranslationData':
		mrna = process_data_info['mRNA']
		protein = process_data_info['protein']
		nucleotide_sequence = process_data_info['nucleotide_sequence']
		organelle = process_data_info['organelle']
		translation = process_data_info['translation']
		transl_table = process_data_info['transl_table']
		pseudo = process_data_info['pseudo']
		process_data = getattr(coralme.core.processdata, process_data_type)(id, model, mrna, protein, nucleotide_sequence, organelle, translation, transl_table, pseudo)
	elif process_data_type == 'tRNAData':
		amino_acid = process_data_info['amino_acid']
		rna = process_data_info['RNA']
		codon = process_data_info['codon']
		process_data = getattr(coralme.core.processdata, process_data_type)(id, model, amino_acid, rna, codon)
	elif process_data_type == 'PostTranslationData':
		processed_protein_id = process_data_info['processed_protein_id']
		unprocessed_protein_id = process_data_info['unprocessed_protein_id']
		process_data = getattr(coralme.core.processdata, process_data_type)(id, model, processed_protein_id, unprocessed_protein_id)
	elif process_data_type == 'GenericData':
		component_list = process_data_info['component_list']
		process_data = getattr(coralme.core.processdata, process_data_type)(id, model, component_list)
		# Create reaction from generic process data
		process_data.create_reactions()
	else:
		process_data = getattr(coralme.core.processdata, process_data_type)(id, model)

	# Set all of the required attributes using information in info dictionary
	for attribute in _REQUIRED_PROCESS_DATA_ATTRIBUTES:
		setattr(process_data, attribute, process_data_dict[attribute])

	# Some attributes depend on process data type. Set those here.
	for attribute in _PROCESS_DATA_TYPE_DEPENDENCIES.get(process_data_type, []):
		value = process_data_info[attribute]
		try:
			setattr(process_data, attribute, value)
		except AttributeError:
			# set to the hidden attribute instead
			setattr(process_data, '_' + attribute, value)

def _add_reaction_from_dict(model, reaction_info):
	"""
	Builds reaction instances defined in dictionary, then add it to the
	ME-model being constructed.

	"""
	reaction_type_dict = reaction_info['reaction_type']

	if len(reaction_type_dict) == 1:
		reaction_type = list(reaction_type_dict.keys())[0]
		reaction_obj = getattr(coralme.core.reaction, reaction_type)(reaction_info['id'])
	else:
		raise Exception('Only 1 reaction_type in valid json')

	for attribute in _REQUIRED_REACTION_ATTRIBUTES:
		# Metabolites are added to reactions using their update function,
		# skip setting metabolite stoichiometries here
		if attribute == 'metabolites':
			continue

		# upper and lower bounds may contain mu values. Handle that here
		value = reaction_info[attribute]
		if attribute in ['upper_bound', 'lower_bound']:
			value = get_sympy_expression(value, model.global_info['growth_key'])
		setattr(reaction_obj, attribute, value)

	# Some reactions are added to model when ME-models are initialized
	try:
		model.add_reactions([reaction_obj])
	except Exception:
		reaction_obj = model.reactions.get_by_id(reaction_obj.id)
		if reaction_type not in ['SummaryVariable', 'GenericFormationReaction'] and not reaction_obj.id.startswith('DM_'):
			warn('Reaction ({:s}) already in model'.format(reaction_obj.id))

	# These reactions types do not have update functions and need their
	# stoichiometries set explicitly .
	if reaction_type in ['SummaryVariable', 'MEReaction']:
		for key, value in reaction_info['metabolites'].items():
			reaction_obj.add_metabolites({model.metabolites.get_by_id(key): get_sympy_expression(value, model.global_info['growth_key'])}, combine=False)

	for attribute in _REACTION_TYPE_DEPENDENCIES.get(reaction_type, []):
		# Spontaneous reactions do no require complex_data
		if attribute == 'complex_data' and 'SPONT' in reaction_obj.id:
			continue

		value = reaction_type_dict[reaction_type][attribute]
		setattr(reaction_obj, attribute, value)

	if hasattr(reaction_obj, 'update'):
		reaction_obj.update()

def me_model_from_dict(obj):
	"""
	Load ME-model from its dictionary representation. This will return
	a full :class:`~coralme.core.model.MEModel` object identical to the
	one saved.

	Parameters
	----------
	obj : dict
		Dictionary representation of ME-model

	Returns
	-------
	:class:`~coralme.core.model.MEModel`:
		Full COBRAme ME-model
	"""

	model = coralme.core.model.MEModel(mu = obj['global_info']['growth_key'])

	for k, v in obj.items():
		if k in {'id', 'name', 'global_info'}:
			setattr(model, k, v)

	# If the ME-model was saved with coralME v1.0,
	# it is missing the default_parameters from the json file.
	# Also, the for-loop will set 'global_info' using the json file,
	# overwriting global_info set by MEModel.__init__()
	if not 'default_parameters' in model.global_info:
		model.default_parameters = {
			'kt' : obj['global_info']['kt'],
			'r0' : obj['global_info']['r0'],
			'k_deg' : obj['global_info']['k_deg'],
			'm_rr' : obj['global_info']['m_rr'],
			'm_aa' : obj['global_info']['m_aa'],
			'm_nt' : obj['global_info']['m_nt'],
			'f_rRNA' : obj['global_info']['f_rRNA'],
			'f_mRNA' : obj['global_info']['f_mRNA'],
			'f_tRNA' : obj['global_info']['f_tRNA'],
			'm_tRNA' : obj['global_info']['m_tRNA'],
			'kcat' : 65.0, # not stored in json with coralME v1.0
			'temperature' : obj['global_info']['temperature'],
			'propensity_scaling' : obj['global_info']['propensity_scaling']
			}
	else:
		model.default_parameters = {
			'kt' : obj['global_info']['default_parameters']['k_t'],
			'r0' : obj['global_info']['default_parameters']['r_0'],
			'k_deg' : obj['global_info']['default_parameters']['k^mRNA_deg'],
			'm_rr' : obj['global_info']['default_parameters']['m_rr'],
			'm_aa' : obj['global_info']['default_parameters']['m_aa'],
			'm_nt' : obj['global_info']['default_parameters']['m_nt'],
			'f_rRNA' : obj['global_info']['default_parameters']['f_rRNA'],
			'f_mRNA' : obj['global_info']['default_parameters']['f_mRNA'],
			'f_tRNA' : obj['global_info']['default_parameters']['f_tRNA'],
			'm_tRNA' : obj['global_info']['default_parameters']['m_tRNA'],
			'kcat' : obj['global_info']['default_parameters']['k^default_cat'],
			'temperature' : obj['global_info']['default_parameters']['temperature'],
			'propensity_scaling' : obj['global_info']['default_parameters']['propensity_scaling']
			}

	model.global_info['growth_key'] = sympy.Symbol(model.global_info['growth_key'], positive = True)

	for metabolite in tqdm.tqdm(obj['metabolites'], 'Adding Metabolites into the ME-model...', bar_format = bar_format):
		_add_metabolite_from_dict(model, metabolite)

	for process_data in tqdm.tqdm(obj['process_data'], 'Adding ProcessData into the ME-model...', bar_format = bar_format):
		_add_process_data_from_dict(model, process_data)

	for reaction in tqdm.tqdm(obj['reactions'], 'Adding Reactions into the ME-model...', bar_format = bar_format):
		_add_reaction_from_dict(model, reaction)

	coralme.builder.compartments.add_compartments_to_model(model)
	model.update()

	return model



#!/usr/bin/python3
import re
import pandas
import logging

class Homology(object):
	"""
	Homology class for storing information about homology of the
	main and reference organisms.

	This class contains methods to predict and process homology
	of the main and reference organisms. Homology is inferred from
	the reciprocal best hits of a BLAST. The results are used to
	update and complement the attributes of the class Organism.

	Parameters
	----------
	org : str
		Identifier of the main organism. Has to be the same as its
		containing folder name.

	ref : str
		Identifier of the reference organism. Has to be the same as
		its containing folder name.

	evalue : float
		E-value cutoff to call enzyme homologs from the BLAST. Two
		reciprocal best hits are considered homologs if their
		E-value is less than this parameter.
	"""

	def __init__(self, org, ref, evalue = False, verbose = False):
		self.org = org
		self.ref = ref

		column_names = 'qseqid\tsseqid\tpident\tlength\tmismatch\tgapopen\tqstart\tqend\tsstart\tsend\tevalue\tbitscore'.split('\t')
		self.org_df = pandas.read_csv(self.org.blast_directory + '/org_as_db.txt', sep = '\t', index_col = 0, names = column_names)
		self.ref_df = pandas.read_csv(self.org.blast_directory + '/ref_as_db.txt', sep = '\t', index_col = 0, names = column_names)

		self.get_mutual_hits(evalue = evalue)

	def get_mutual_hits(self, evalue = False, verbose = False):
		# to not import from coralme.builder.homology
		def get_top_hits(df, evalue = False):
			top_hits = {}
			for g in df.index.unique():
				hits = df.loc[g]
				if isinstance(hits, pandas.DataFrame):
					top_hit = hits.sort_values(by = 'evalue').iloc[0]
				else:
					top_hit = hits
				if evalue and top_hit['evalue'] > evalue:
					continue
				top_hits[g] = top_hit
			return top_hits

		org_df = self.org_df
		ref_df = self.ref_df

		if verbose:
			logging.warning('Getting top hits...')

		org_top_hits = get_top_hits(org_df, evalue = evalue)
		ref_top_hits = get_top_hits(ref_df, evalue = evalue)

		if verbose:
			logging.warning('Getting reciprocal top hits...')

		mutual_hits = {}
		for g, hit in org_top_hits.items():
			hit_g = hit['sseqid']

			if hit_g not in ref_top_hits:
				continue

			if g == ref_top_hits[hit_g]['sseqid']:
				mutual_hits[g] = {}
				mutual_hits[hit_g] = {}

				mutual_hits[g]['query'] = hit_g
				mutual_hits[g]['evalue'] = hit['evalue']

				mutual_hits[hit_g]['query'] = g
				mutual_hits[hit_g]['evalue'] = ref_top_hits[hit_g]['evalue']

		if verbose:
			logging.warning('Done')

		mutual_hits_df = pandas.DataFrame.from_dict(mutual_hits).T.sort_index()
		self.mutual_hits_df = mutual_hits_df
		self.mutual_hits = mutual_hits_df['query'].to_dict()

	def get_complex_homology(self):
		org_complexes_df = self.org.complexes_df
		ref_complexes_df = self.ref.complexes_df
		mutual_hits = self.mutual_hits

		# Get ref homology for complexes already annotated in BioCyc
		org_cplx_homolog = {}
		ref_cplx_homolog = {}
		not_annotated_candidates = set()
		warn_candidates = []

		for c, row in org_complexes_df.iterrows():
			if isinstance(row['genes'], float) or not row['genes']:
				continue

			genes = [re.findall('.*(?=\(\d*\))', g)[0] for g in row['genes'].split(' AND ')]

			if not genes or not set(genes).issubset(set(mutual_hits.keys())):
				continue  # All org genes must have a hit

			for g in genes:
				if c in org_cplx_homolog:
					break

				ref_gene = mutual_hits[g]
				ref_complexes = ref_complexes_df[ref_complexes_df['genes'].str.contains(ref_gene)]
				for rc, rrow in ref_complexes.iterrows():
					rgenes = [re.findall('.*(?=\(\d*\))', g)[0] for g in rrow['genes'].split(' AND ')]
					if not set(rgenes).issubset(set(mutual_hits.keys())):
						continue  # All ref genes must have a hit
					ogenes = [mutual_hits[og] for og in genes]
					if set(rgenes) == set(ogenes):  # Complex identified
						org_cplx_homolog[c] = rc
						ref_cplx_homolog[rc] = c
					else:
						if rc not in not_annotated_candidates:
							warn_candidates.append({
								'complex': c,
								'reference_complex' : rc
								})
						not_annotated_candidates.add(rc)

		not_annotated_candidates = not_annotated_candidates.difference(set(ref_cplx_homolog.keys()))
		logging.warning('{} complexes were mapped successfully'.format(len(org_cplx_homolog)))

		if warn_candidates:
			self.org.curation_notes['org.get_complex_homology'].append({
				'msg':'Some complexes were partial hits in the BLAST',
				'triggered_by':warn_candidates,
				'importance':'medium',
				'to_do':'Curate these manually in protein_corrections.txt'})
		for i in self.org.generic_dict.keys():
			org_cplx_homolog[i] = i
			ref_cplx_homolog[i] = i
		self.org_cplx_homolog = org_cplx_homolog
		self.ref_cplx_homolog = ref_cplx_homolog
		self.not_annotated_candidates = not_annotated_candidates



import tqdm
bar_format = '{desc:<75}: {percentage:.1f}%|{bar:10}| {n_fmt:>5}/{total_fmt:>5} [{elapsed}<{remaining}]'
import numpy
import pandas

import logging
log = logging.getLogger(__name__)

import coralme

# read genbank and extract sequences and data
import Bio
from Bio import SeqIO, Seq, SeqFeature, SeqUtils

def add_transcription_reaction(me_model, tu_name, locus_ids, sequence, organelle = None, update = True):
	"""
	Create TranscriptionReaction object and add it to ME-model.
	This includes the necessary transcription data.

	Parameters
	----------
	me_model : :class:`coralme.core.model.MEModel`
		The MEModel object to which the reaction will be added

	tu_name : str
		ID of TU being transcribed.
		The TranscriptionReaction will be added as "transcription_+TU_name"
		The TranscriptionData will be added as just 'TU_name'

	locus_ids : set
		Set of locus IDs that the TU transcribes

	sequence : str
		Nucleotide sequence of the TU.

	update : bool
		If True, use TranscriptionReaction's update function to update and
		add reaction stoichiometry

	Returns
	-------
	:class:`coralme.core.reaction.TranscriptionReaction`
		TranscriptionReaction for the TU
	"""

	# 1) Create TranscriptionReaction
	transcription = coralme.core.reaction.TranscriptionReaction('transcription_' + tu_name)

	# 2) add TranscriptionData into TranscriptionReaction
	transcription.transcription_data = coralme.core.processdata.TranscriptionData(
		id = tu_name, model = me_model, nucleotide_sequence = sequence, rnap = '',
		rna_products = {'RNA_' + i for i in locus_ids}, organelle = organelle
		)

	# 3) and add TranscriptionReaction into ME-Model
	me_model.add_reactions([transcription])
	logging.warning('A TranscriptionReaction with ID \'transcription_{:s}\' was added to the ME-model.'.format(tu_name))

	if update:
		transcription.update()

def create_transcribed_gene(me_model, locus_id, rna_type, seq, left_pos = None, right_pos = None, strand = None):
	"""
	 Creates a `TranscribedGene` metabolite object and adds it to the ME-model

	Parameters
	----------
	me_model : :class:`coralme.core.model.MEModel`
		The MEModel object to which the reaction will be added

	locus_id : str
		Locus ID of RNA product.
		The TranscribedGene will be added as "RNA + _ + locus_id"

	left_pos : int or None
		Left position of gene on the sequence of the (+) strain

	right_pos : int or None
		Right position of gene on the sequence of the (+) strain

	seq : str
		Nucleotide sequence of RNA product.
		Amino acid sequence, codon counts, etc. will be calculated based on
		this string.

	strand : str or None
		- (+) if the RNA product is on the leading strand
		- (-) if the RNA product is on the complementary strand

	rna_type : str
		Type of RNA of the product.
		tRNA, rRNA, or mRNA
		Used for determining how RNA product will be processed.

	Returns
	-------
		:class:`coralme.core.component.TranscribedGene`
			Metabolite object for the RNA product
	"""
	gene = coralme.core.component.TranscribedGene('RNA_' + locus_id, rna_type, seq)
	fn = lambda x: int(x.replace('>', '').replace('<', ''))
	gene.left_pos = sorted(left_pos.split(','), key = fn) if left_pos is not None else None
	gene.right_pos = sorted(right_pos.split(','), key = fn) if right_pos is not None else None
	gene.strand = strand

	if len(me_model.metabolites.query('^RNA_{:s}$'.format(locus_id))) != 0:
		me_model.metabolites._replace_on_id(gene)
		logging.warning('A Metabolite component with ID \'RNA_{:s}\' was replaced with a TranscribedGene component \'{:s}\'.'.format(locus_id, gene.id))
	else:
		me_model.add_metabolites([gene])
		logging.warning('A TranscribedGene component with ID \'RNA_{:s}\' was added to the ME-model.'.format(locus_id))

def add_translation_reaction(me_model, locus_id, dna_sequence, prot_sequence = '', organelle = None, transl_table = 1, pseudo = False, update = False):
	"""
	Creates and adds a TranslationReaction to the ME-model as well as the
	associated TranslationData

	A dna_sequence is required in order to add a TranslationReaction to the
	ME-model

	Parameters
	----------
	me_model : :class:`coralme.core.model.MEModel`
		The MEModel object to which the reaction will be added

	locus_id : str
		Locus ID of RNA product.
		The TranslationReaction will be added as "translation + _ + locus_id"
		The TranslationData will be added as 'locus_id'

	dna_sequence : str
		DNA sequence of the RNA product. This string should be reverse
		transcribed if it originates on the complement strand.

	update : bool
		If True, use TranslationReaction's update function to update and
		add reaction stoichiometry

	"""
	# Add RNA to model if it doesn't exist
	if 'RNA_' + locus_id not in me_model.metabolites:
		rna = coralme.core.component.TranscribedGene('RNA_' + locus_id, 'mRNA', dna_sequence)
		logging.warning('The \'RNA_{:s}\' component was not present in ME-model and it was created.'.format(locus_id))
		me_model.add_metabolites(rna)

	# 1) Create TranslationReaction
	translation_reaction = coralme.core.reaction.TranslationReaction('translation_' + locus_id)

	# 2) add TranslationData into TranslationReaction
	translation_reaction.translation_data = coralme.core.processdata.TranslationData(
		id = locus_id, model = me_model, mrna = 'RNA_' + locus_id, protein = 'protein_' + locus_id,
		nucleotide_sequence = dna_sequence, organelle = organelle, translation = prot_sequence,
		transl_table = Bio.Data.CodonTable.generic_by_id[transl_table], pseudo = pseudo
		)

	# 3) and add TranslationReaction into ME-Model
	me_model.add_reactions([translation_reaction])
	logging.warning('A TranslationReaction with ID \'translation_{:s}\' was added to the ME-model.'.format(locus_id))

	if update:
		translation_reaction.update()

def convert_aa_codes_and_add_charging(me_model, trna_to_aa, trna_to_codon, organelle, verbose = True):
	"""
	Adds tRNA charging reactions for all tRNAs in ME-model

	Parameters
	----------
	me_model : :class:`coralme.core.model.MEModel`
		The MEModel object to which the reaction will be added

	trna_to_aa : dict
		Dictionary of tRNA locus ID to 3 letter codes of the amino acid
		that the tRNA contributes

		{tRNA identifier (locus_id): amino_acid_3_letter_code}

	trna_to_codon : dict
		Dictionary of tRNA identifier to the codon which it associates

		{tRNA identifier (locus_id): codon_sequence}

	verbose : bool
		If True, display metabolites that were not previously added to the
		model and were thus added when creating charging reactions
	"""

	# remove "other" tRNAs: This avoids a RuntimeError: dictionary changed size during iteration using pop
	trna_to_aa = { k:v for k,v in trna_to_aa.items() if v.lower() != 'other' }

	# update trna_to_aa and trna_to_codon dictionaries with genetic_recoding
	if me_model.global_info['genetic_recoding']:
		for recoded_codon, recoded_aa_to_tRNA in me_model.global_info['genetic_recoding'].items():
			for aa, tRNA in recoded_aa_to_tRNA.items():
				trna_to_aa.update({ tRNA : aa.replace('__L_c', '') })
				trna_to_codon[tRNA] = [recoded_codon] if isinstance(recoded_codon, str) else recoded_codon

	# convert amino acid 3 letter codes to metabolites
	#for tRNA, aa in list(iteritems(trna_to_aa)):
	for tRNA, aa in trna_to_aa.items():
		if aa.lower() == 'other':
			pass
			#trna_to_aa.pop(tRNA) # RuntimeError: dictionary changed size during iteration
		elif aa.lower() == 'sec':
			# Charge with precursor to selenocysteine
			trna_to_aa[tRNA] = me_model.metabolites.get_by_id('ser__L_' + organelle)
		elif aa.lower() == 'gly':
			trna_to_aa[tRNA] = me_model.metabolites.get_by_id('gly_' + organelle)
		else:
			if me_model.metabolites.has_id(aa.lower() + '__L_' + organelle):
				trna_to_aa[tRNA] = me_model.metabolites.get_by_id(aa.lower() + '__L_' + organelle)
			else:
				logging.warning('The amino acid \'{:s}\' does not exist in the ME-model.'.format(aa.lower() + '__L_' + organelle))

	# check trna_to_codon for START codon
	#start = False
	#for tRNA, codon in trna_to_codon[organelle].items():
		#if 'START' in codon:
			#start = True
	if not any([ True if 'START' in v else False for k,v in trna_to_codon.items() ]):
		logging.warning('Associate at least one tRNA-Met/tRNA-fMet gene with the \'START\' keyword or add manually a \'tRNAChargingReaction\'.')

	# add in all the tRNA charging reactions
	for tRNA, aa in trna_to_aa.items():
		# trna_to_codon contains all the tRNAs found in the organism, not a subset of tRNAs per organelle
		if tRNA not in trna_to_codon.keys() or isinstance(aa, str):
			continue

		if not me_model.metabolites.has_id(aa.id):
			continue

		for codon in trna_to_codon[tRNA]:
			codon = codon.replace('T', 'U') if codon != 'START' else codon
			trna_data = coralme.core.processdata.tRNAData('tRNA_{:s}_{:s}'.format(tRNA, codon), me_model, aa.id, 'RNA_' + tRNA, codon)
			charging_reaction = coralme.core.reaction.tRNAChargingReaction('charging_tRNA_{:s}_{:s}'.format(tRNA, codon))
			charging_reaction.tRNA_data = trna_data
			charging_reaction.organelle = organelle

			me_model.add_reactions([charging_reaction])
			charging_reaction.update(verbose = verbose)
			logging.warning('A tRNAChargingReaction with ID \'tRNA_{:s}_{:s}\' was added to the ME-model.'.format(tRNA, codon))

def build_reactions_from_genbank(
	me_model, gb_filename, tu_frame = pandas.DataFrame(columns = ['genes']), genes_to_add = list(),
	feature_types = [ 'CDS', 'rRNA', 'tRNA', 'ncRNA', 'tmRNA', 'misc_RNA' ], update = True, verbose = True,
	trna_misacylation = dict(), genome_mods = dict(), knockouts = list()):
	# trna_to_codon = dict(), frameshift_dict = None, # not needed anymore

	"""Creates and adds transcription and translation reactions using genomic
	 information from the organism's genbank file. Adds in the basic
	 requirements for these reactions. Organism specific components are added.

	Parameters
	----------
	me_model : :class:`coralme.core.model.MEModel`
		The MEModel object to which the reaction will be added

	gb_filename : str
		Local name of the genbank file that will be used for ME-model
		construction

	tu_frame : :class:`pandas.DataFrame`
		DataFrame with indexes of the transcription unit name and columns
		containing the transcription unit starting and stopping location on
		the genome and whether the transcription unit is found on the
		main (+) strand or complementary (-) strand.

		If no transcription unit DataFrame is passed into the function,
		transcription units are added corresponding to each transcribed
		gene in the genbank file.

	element_types : set
		Transcription reactions will be added to the ME-model for all RNA
		feature.types in this set. This uses the nomenclature of the
		genbank file (gb_filename)

	verbose : bool
		If True, display metabolites that were not previously added to the
		model and were thus added when creating charging reactions
	"""
	# old docstring
	#frameshift_dict : dict
		#{locus_id: genome_position_of_TU}

		#If a locus_id is in the frameshift_dict, update it's nucleotide
		#sequence to account of the frameshift

	#if not trna_to_codon:
		#trna_to_codon = {}

	#metabolites = me_model.metabolites

	# Load genbank file and extract DNA sequence
	#try:
		#seqs = SeqIO.read(gb_filename, 'gb')
		#full_seqs = str(seqs.seq)
	#except:
	contigs = []
	if isinstance(gb_filename, str):
		gb_filename = [gb_filename]
	for infile in gb_filename:
		for contig in SeqIO.parse(infile, 'gb'):
			contigs.append(contig)
	full_seqs = { x.id:x.seq for x in contigs }
	me_model.global_info['full_seqs'] = full_seqs

	# GC Content
	if me_model.global_info.get('GC_fraction', None) is None:
		# Deprecated in Biopython 1.80
		#me_model.global_info['GC_fraction'] = SeqUtils.GC(''.join([ str(x) for x in full_seqs.values()]))
		me_model.global_info['GC_fraction'] = SeqUtils.gc_fraction(''.join([ str(x) for x in full_seqs.values()]))

	# modify sequence(s) using the genome_mods dictionary
	for replicon, coords in genome_mods.items():
		new = full_seqs[replicon]
		for segment in coords.split(';'):
			if segment.startswith('DEL:'):
				x = segment[4:].split('..')
				new = ''.join([ '-' if idx in range(int(x[0])-1, int(x[1])) else nucl for idx,nucl in enumerate(new) ])
			elif segment.startswith('INS:'):
				raise NotImplementedError
			elif segment[0] in ['A', 'T', 'C', 'G'] and segment[-1] in ['A', 'T', 'C', 'G'] and segment[1:-1].isdigit():
				new[int(segment[1:-1])-1] = segment[-1]
			else:
				x = segment.split(':')
				if len(x[0]) == len(x[1]):
					new = new.replace(x[0], x[1]) # whole-genome recoding
				elif len(x[0]) >= len(x[1]):
					x[1] == x[1] + '-' * (len(x[0]) - len(x[1]))
					new = new.replace(x[0], x[1]) # whole-genome recoding
					logging.warning('Genome modification involved the deletion of nucleotides and we proceeded as instructed.')
				else:
					logging.warning('Genome modification involves the insertion of nucleotides and we won\'t proceed as instructed.')
					raise NotImplementedError

		full_seqs[replicon] = Seq.Seq(new)

	# copy new sequences back
	for contig in contigs:
		contig.seq = full_seqs[contig.id]

	# If no tu_frame is provided generate a new TU frame where each mRNA gets its own TU
	#using_tus = tu_frame is not None
	#if not using_tus:
		#tu_frame = pandas.DataFrame(columns = ['genes'])
		#tu_frame = []
		#for contig in gb_file:
			#tu_frame.append(pandas.DataFrame.from_dict({
				#'TU_' + feature.qualifiers['locus_tag'][0]: {
					##'start': int(feature.location.start),
					##'stop': int(feature.location.end),
					#'replicon': contig.id,
					#'genes': dict(feature.qualifiers)['locus_tag'],
					#'start': ','.join([ str(x.start) for x in feature.location.parts ]),
					#'stop': ','.join([ str(x.end) for x in feature.location.parts ]),
					#'strand': "+" if feature.strand == 1 else "-",
					#} for feature in contig.features if (feature.type in element_types and feature.qualifiers.get('pseudo') is None)
				#}, orient = 'index'))
		#tu_frame = pandas.concat(tu_frame, axis = 0)

	# Create transcription reactions for each TU and DNA sequence.
	# RNA_products will be added so no need to update now
	for tu_id in tqdm.tqdm(tu_frame.index, 'Adding Transcriptional Units into the ME-model from user input...', bar_format = bar_format):
		# in rare cases, transcription units have no genes associated to them
		if tu_frame.genes[tu_id] == '': # we read df_tus as strings
			logging.warning('The transcription unit \'{:s}\' has no genes associated to it. Please check if it is the correct behavior.'.format(tu_id))
			continue

		if any(x in tu_frame.genes[tu_id].split(',') for x in genes_to_add):
			start = tu_frame.start[tu_id]
			stop = tu_frame.stop[tu_id]
			strand = 1 if tu_frame.strand[tu_id] == '+' else -1
			organelle = tu_frame.organelle[tu_id] if 'organelle' in tu_frame.columns else None

			#if len(str(start).split(',')) > 1:
				#locations = []
				#for start, stop in zip(start.split(','), stop.split(',')):
					#locations.append(SeqFeature.FeatureLocation(
						#SeqFeature.ExactPosition(int(start)-1), SeqFeature.ExactPosition(int(stop)), strand = strand))

				#seq = SeqFeature.SeqFeature(SeqFeature.CompoundLocation(locations, 'join'))
			#else:
				#seq = SeqFeature.SeqFeature(SeqFeature.FeatureLocation(
					#SeqFeature.ExactPosition(int(start)-1), SeqFeature.ExactPosition(int(stop)), strand = strand))

			seqfeatures = []
			#for start, stop in zip(starts.split(','), stops.split(',')):
			if int(start) < int(stop):
				# nicely defined locus in reference of the genome sequence
				seqfeatures.append([SeqFeature.SeqFeature(SeqFeature.SimpleLocation(int(start)-1, int(stop)))])
			else:
				# the feature must be split in two
				loc1 = SeqFeature.SeqFeature(SeqFeature.SimpleLocation(int(start)-1, len(full_seqs[tu_frame.replicon[tu_id]])))
				loc2 = SeqFeature.SeqFeature(SeqFeature.SimpleLocation(0, int(stop)))
				seqfeatures.append([loc1, loc2])

			#sequence = coralme.util.dogma.extract_sequence(
				#full_seqs[tu_frame.replicon[tu_id]],
				#tu_frame.start[tu_id],
				#tu_frame.stop[tu_id],
				#tu_frame.strand[tu_id],
				#)

			#replicons = tu_frame.genes[tu_id] if tu_frame.replicon[tu_id] is numpy.nan else tu_frame.replicon[tu_id]
			replicons = tu_frame.replicon[tu_id].split(',')
			#print(tu_id, replicons)
			dna = ''
			for replicon_id, seqs in zip(replicons, seqfeatures):
				#print(replicon_id, seq)
				# 'ungap' method and 'strand' argument are deprecated in Biopython 1.80
				#seq = seq.extract(full_seqs[tu_frame.replicon[tu_id]]).ungap()
				for seq in seqs:
					seq.strand = strand
					dna += seq.extract(full_seqs[replicon_id]).replace('-', '')

			if len(dna) == 0:
				logging.warning('The knockouts dictionary instructed to completely delete \'{:s}\' from the ME-model.'.format(tu_id))
			else:
				add_transcription_reaction(me_model, tu_id, set(), str(dna), organelle, update = False)

	canonical_aas = [
		'Ala', 'Arg', 'Asn', 'Asp', 'Cys', 'Gln', 'Glu', 'Gly', 'His', 'Ile',
		'Leu', 'Lys', 'Met', 'Phe', 'Pro', 'Ser', 'Thr', 'Trp', 'Tyr', 'Val'
		]

	# Dictionary of tRNA locus ID to the model.metabolite object. It accounts for misacylation
	trna_to_aa = {}

	# Dictionary of tRNA locus ID to amino acid, one dict of tRNAs per organelle type
	# aa2trna does not account for misacylation
	aa2trna = { 'c' : {} } # prokaryotes and eukaryotes
	if me_model.global_info['domain'].lower() not in ['prokaryote', 'bacteria']:
		aa2trna.update({'m' : {}, 'h' : {}}) # mitochondria and plastids

	# Translation tables, one table per organelle type
	transl_tables = { 'c' : set() } # prokaryotes and eukaryotes
	if me_model.global_info['domain'].lower() not in ['prokaryote', 'bacteria']:
		transl_tables.update({'m' : set(), 'h' : set()}) # mitochondria and plastids

	# Set of start and stop codons
	start_codons = set()
	stop_codons = set()

	# Codon usage
	from collections import Counter
	codon_usage = Counter()

	# New Gene Locus ID
	new_locus_tag_counter = 1

	# Associate each feature (RNA_product) with a TU and add translation reactions and demands
	for contig in contigs:
		iterator = tqdm.tqdm(contig.features, 'Adding features from contig {:s} into the ME-model...'.format(contig.id), bar_format = bar_format) if len(contigs) < 10 else contig.features
		for feature in iterator:
			# Find organelle in source
			if feature.type in ['source', 'region']: # genbanks from gff+fna do not contain a 'source' feature
				organelle = feature.qualifiers.get('organelle', [None])[0]
				continue

			# Optionally add pseudo genes into the ME-model
			if not me_model.global_info['include_pseudo_genes'] and 'pseudo' in feature.qualifiers:
				logging.warning('The feature \'{:s}\' is a pseudogene. Use \'"include_pseudo_genes" : True\' to add the feature into the model.'.format(bnum))
				continue

			# Add only features based on their type
			if feature.type not in feature_types:
				continue

			# Some features might lack a locus tag
			if not feature.qualifiers.get(me_model.global_info.get('locus_tag', 'locus_tag'), False):
				logging.warning('The feature \'{:s}\' of type \'{:s}\', located at \'{:s}\' misses a \'{:s}\' qualifier.'.format(feature.qualifiers.get('product', ['no product name'])[0], feature.type, str(feature.location), me_model.global_info.get('locus_tag', 'locus_tag')))
				filter1 = feature.qualifiers.get('gene', ['no product name'])[0].startswith('tRNA-')
				filter2 = feature.qualifiers.get('product', ['no product name'])[0].startswith('tRNA-')
				if filter1 or filter2:
					feature.qualifiers[me_model.global_info.get('locus_tag', 'locus_tag')] = new_locus_tag = ['CORALME_{:03d}'.format(new_locus_tag_counter)]
					logging.warning('The feature was identified as a tRNA and assigned the Gene Locus ID \'{:s}\'.'.format(new_locus_tag[0]))
					new_locus_tag_counter += 1
				else:
					logging.warning('The gene identified will be ignored from the reconstruction.')
					continue

			bnum = feature.qualifiers[me_model.global_info.get('locus_tag', 'locus_tag')][0]

			if me_model.process_data.has_id(bnum):
				logging.warning('A gene with a Gene Locus ID \'{:s}\' was added previously. Please, check the GenBank file and correct it accordingly.'.format(bnum))
				continue

			# Skip feature if it is not a gene used in the ME-model reconstruction
			filter1 = bnum.startswith('CORALME_')
			filter2 = bnum in knockouts
			filter3 = bnum not in genes_to_add
			if filter1:
				pass
			elif filter2 or filter3:
				logging.warning('The genomic feature \'{:s}\' is in the knockouts list.'.format(bnum))
				continue

			if feature.type in [ 'ncRNA', 'tmRNA', 'misc_RNA', 'RNA' ]:
				# list of rna components
				me_model.global_info['rna_components'].append(bnum)

			# Assign values for all important gene attributes
			# old code cannot consider if genes are split
			#left_pos = int(feature.location.start)
			#right_pos = int(feature.location.end)
			left_pos = ','.join([ str(x.start) for x in feature.location.parts ])
			right_pos = ','.join([ str(x.end) for x in feature.location.parts ])
			rna_type = 'mRNA' if feature.type == 'CDS' else feature.type
			strand = '+' if feature.strand == 1 else '-'
			# old code uses a cannon to hit a nail
			#seq = coralme.util.dogma.extract_sequence(full_seqs[contig.id], left_pos, right_pos, strand)

			# Deprecated in Biopython 1.80
			#seq = feature.extract(contig).seq.ungap() # using Biopython is better
			seq = feature.extract(contig).seq.replace('-', '')
			if len(seq) == 0:
				logging.warning('The genomic feature \'{:s}\' sequence is null. Please check GenBank file coordinates and sequence.'.format(bnum))
				continue

			# old code uses a dictionary setting the frameshifts.
			# the genbank already include frameshifts in the location of features
			# the genbank also sets the strand in the location; no need to reverse_transcribe()
			## Add gene metabolites and apply frameshift mutations----
			#frameshift_string = frameshift_dict.get(bnum)
			#if len(seq) % 3 != 0 and frameshift_string:
				#print('Applying frameshift on {:s}'.format(bnum))
				#seq = coralme.util.dogma.return_frameshift_sequence(full_seq, frameshift_string)
				#if strand == '-':
					#seq = coralme.util.dogma.reverse_transcribe(seq)

			# Add TranscribedGene metabolite
			create_transcribed_gene(me_model, bnum, rna_type, str(seq), left_pos, right_pos, strand)

			# Add translation reaction for mRNA
			# builder.generate_files will create a modified genbank
			# If the user runs a configuration with the original genbank, pseudogenes could not assert len(seq) % 3 == 0
			# Also, genomic modification that are not paired correctly within the knockouts list could not assert len(seq) % 3 == 0
			#if rna_type == 'mRNA' and len(seq) % 3 == 0:
			if rna_type == 'mRNA':
				# Add the translation table
				#prot = feature.qualifiers.get('translation', [''])[0] # get translation from DNA sequence
				transl_table = feature.qualifiers.get('transl_table', ['1'])[0]
				prot = seq.translate(transl_table)
				pseudo = feature.qualifiers.get('pseudo', False)
				add_translation_reaction(me_model, bnum, dna_sequence = str(seq), prot_sequence = str(prot), organelle = organelle, transl_table = int(transl_table), pseudo = bool(pseudo), update = False)

				# Add the start codon to the start_codons set
				start_codons.add(str(seq[:3]).replace('T', 'U'))
				# Add the stop codon to the stop_codons sets
				stop_codons.add(str(seq[-3:]).replace('T', 'U'))

				# Add the translation table per organelle
				if organelle is None:
					transl_tables['c'].add(int(transl_table))
					if me_model.global_info['domain'].lower() not in ['prokaryote', 'bacteria']:
						logging.warning('Contig \'{:s}\' does not report an organelle type.'.format(contig.id))
				elif organelle.lower() in ['mitochondria', 'mitochondrion']:
					transl_tables['m'].add(int(transl_table))
				elif organelle.lower() in ['chloroplast', 'plastid']:
					transl_tables['h'].add(int(transl_table))
				else:
					continue

				# Add the codon usage
				if len(seq) % 3 == 0:
					codons = [ str(seq[pos:pos+3]) for pos in range(0, len(seq), 3) ]
					codon_usage.update(Counter(codons))
				else:
					logging.warning('Gene \'{:s}\' was not used to determine the codon usage.'.format(bnum))

			me_model.global_info['transl_tables'] = transl_tables
			me_model.global_info['start_codons'] = start_codons
			me_model.global_info['stop_codons'] = stop_codons
			me_model.global_info['codon_usage'] = codon_usage

			## Associate the TranscribedGene to TU(s)
			# old code does not consider that a gene can start at the "end" of the genome and finish at the "start" of it
			#parent_tu = tu_frame[(tu_frame.start - 1 <= left_pos) & (tu_frame.stop >= right_pos) & (tu_frame.strand == strand)].index
			parent_tu = tu_frame[[ True if bnum in x.split(',') else False for x in tu_frame.genes.values ]].index.tolist()

			if len(parent_tu) == 0:
				tu_id = 'TU_' + bnum
				parent_tu = [tu_id]
				add_transcription_reaction(me_model, tu_id, set(), str(seq), organelle, update = False)
				logging.warning('No Trancriptional Unit found for {:s} {:s}. Created a dummy TU_{:s} component.'.format(rna_type, bnum, bnum))

			for TU_id in parent_tu:
				me_model.process_data.get_by_id(TU_id).RNA_products.add('RNA_' + bnum)
				me_model.process_data.get_by_id(TU_id).original_RNA_products.add('RNA_' + bnum)

			# Deal with the complicated tRNA biology
			# me_model.global_info['trna_to_codon'] sub-dictionaries are created empty during MEModel.__init__; completed during generate_files()
			if me_model.global_info['trna_to_codon'].get(organelle, {}) != {}:
				continue
			else:
				# Create dict to use for adding tRNAChargingReactions later
				# tRNA_aa = {'tRNA':'amino_acid'}
				msg1 = 'From the tRNA misacylation dictionary, the {:s} gene [tRNA({:s})] is loaded and converted into {:s}-tRNA({:s}). Make sure a MetabolicReaction to convert a {:s}-tRNA({:s}) into a {:s}-tRNA({:s}) is present in the ME-model.'
				msg2 = 'From the tRNA misacylation dictionary, the {:s} gene [tRNA({:s})] is loaded and converted into {:s}-tRNA({:s}). No further modification needs to take place.'

				if rna_type == 'tRNA':
					aa = feature.qualifiers.get('product', ['tRNA-None'])[0].split('-')[1]
					if aa in canonical_aas + ['Asx', 'Glx', 'fMet', 'Sec']:
						pass
					else:
						logging.warning('The tRNA \'{:s}\' is not associated to a valid product name (tRNA-Amino acid 3 letters code)'.format(bnum))
						continue

					#aa2trna[bnum] = aa # original tRNA<->Amino acid association to be used later in trna_to_codon

					msg = 'The tRNA \'{:s}\' is associated to two amino acids. The \'trna_misacylation\' dictionary was modified to attempt load the correct amino acid.'
					# Special tRNA(Asx) that can be loaded with Asn (EC 6.1.1.22) or Asp (EC 6.1.1.12)
					# If loaded with Asp, it is converted into Asn (EC 6.3.5.6)
					if aa == 'Asx':
						trna_misacylation['Asx'] = 'Asp'
						logging.warning(msg.format(bnum))
					# Special tRNA(Glx) that can be loaded with Gln (EC 6.1.1.18) or Glu (EC 6.1.1.17)
					# If loaded with Glu, it is converted into Gln (EC 6.3.5.7)
					if aa == 'Glx':
						trna_misacylation['Glx'] = 'Glu'
						logging.warning(msg.format(bnum))

					me_model.global_info['trna_misacylation'] = trna_misacylation

					# misacylation of glutamate/aspartate occurs in archaea, Gram-positive eubacteria, mitochondria, and chloroplasts
					if aa in trna_misacylation.keys():
						# misacylation only in mitochondria and chloroplasts
						filter1a = me_model.global_info['domain'].lower() in ['eukarya', 'eukaryote']
						filter1b = str(organelle).lower() in ['mitochondria', 'mitochondrion', 'chloroplast', 'plastid']
						# misacylation in the cytoplasm of Gram-positive eubacteria (and other bacteria such as cyanobacteria)
						filter2 = me_model.global_info['domain'].lower() in ['bacteria', 'prokaryote']

						if filter1a and filter1b or filter2:
							trna_to_aa[bnum] = trna_misacylation[aa]
							if aa.endswith('x'):
								logging.warning(msg2.format(bnum, aa, trna_misacylation[aa], aa))
								#aa = trna_misacylation[aa]
							else:
								logging.warning(msg1.format(bnum, aa, trna_misacylation[aa], aa, trna_misacylation[aa], aa, aa, aa))
						else:
							# misacylation is not valid in the compartment and domain
							trna_to_aa[bnum] = aa
					else:
						trna_to_aa[bnum] = aa

					if organelle is None:
						aa2trna['c'][bnum] = aa
					elif organelle.lower() in ['mitochondria', 'mitochondrion']:
						aa2trna['m'][bnum] = aa
					elif organelle.lower() in ['chloroplast', 'plastid']:
						aa2trna['h'][bnum] = aa
					#old code
					#trna_to_aa[bnum] = feature.qualifiers["product"][0].split('-')[1]

				# trna_to_codon does not account for misacylation: { 'tRNA ID' : 'Amino acid to load into the tRNA' }
				trna_to_aa = { k:v.replace('fMet', 'Met') for k,v in trna_to_aa.items() }
				me_model.global_info['trna_to_aa'] = trna_to_aa

				# DataFrame mapping tRNAs (list) and the encoded amino acid (index), per organelle
				# aa2trna derives from trna_to_aa, so it also accounts for misacylation: { 'organelle ID' : 'DataFrame of amino acid to load into the tRNA' }
				#me_model.global_info['aa2trna'] = aa2trna

	for organelle, aa2trna_dct in aa2trna.items():
		aa2trna_dct = { k:v.capitalize().split('_')[0] if 'fMet' not in v else 'fMet' for k,v in aa2trna_dct.items() }
		aa2trna_df = pandas.DataFrame(data = [aa2trna_dct.values(), aa2trna_dct.keys()]).T
		aa2trna_df = aa2trna_df.groupby(0).agg({1: lambda x: x.tolist()})
		if 'fMet' in aa2trna_df.index:
			aa2trna_df.loc['Met'] = aa2trna_df.loc['fMet'] + aa2trna_df.loc['Met']

		aa2trna[organelle] = aa2trna_df

		if not aa2trna_df.empty:
			# assign START tRNAs to every fMet-tRNA (Met-tRNA if not) and check if at least one tRNA was identified
			if 'fMet' in aa2trna_df.index:
				if len(me_model.global_info['START_tRNA']) == 0:
					me_model.global_info['START_tRNA'] = list(aa2trna_df.loc['fMet'])[0]

			if 'Met' in aa2trna_df.index:
				if len(me_model.global_info['START_tRNA']) == 0:
					me_model.global_info['START_tRNA'] = list(aa2trna_df.loc['Met'])[0]

			# final check
			if len(me_model.global_info['START_tRNA']) == 0:
				logging.warning('Unable to identify at least one \'tRNA-Met\' or \'tRNA-fMet\' annotation from the \'Definition\' column in the organism-specific matrix.')
		else:
			logging.warning('No tRNA genes were identified from their locus tags.')

		# DataFrame mapping tRNAs (list) and the encoded amino acid (index), per organelle
		# aa2trna derives from trna_to_aa, so it also accounts for misacylation: { 'organelle ID' : 'DataFrame of amino acid to load into the tRNA' }
		me_model.global_info['aa2trna'] = aa2trna

		# add charging tRNA reactions per organelle
		for organelle, transl_table in transl_tables.items():
			if len(transl_table) == 0:
				continue

			#codon_table = Bio.Data.CodonTable.generic_by_id[me_model.global_info.get('translation_table', 11)]
			#me_model.global_info['codon_table'] = codon_table
			codon_table = Bio.Data.CodonTable.generic_by_id[list(transl_table)[0]]

			dct = { k.replace('T', 'U'):SeqUtils.seq3(v) for k,v in codon_table.forward_table.items() if 'U' not in k }
			aa2codons = pandas.DataFrame(data = [dct.keys(), dct.values()]).T.groupby(1).agg({0: lambda x: x.tolist()})
			# aa2codons derives from the translation table and maps amino acids to the codon
			#me_model.global_info['aa2codons'][organelle] = aa2codons

			#if me_model.global_info.get('translation_table', None) is None:
				#me_model.global_info['translation_table'] = int(list(transl_table)[0]) if len(transl_table) == 1 else 11
				#logging.warning('Translation table was set to {:d} (See more https://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi#SG{:d}).'.format(me_model.global_info['translation_table'], me_model.global_info['translation_table']))

			#if me_model.global_info['translation_table'] == 11:
			if list(transl_table)[0] == 11:
				aa2codons.loc['Sec'] = [['UGA']] # an internal UGA encodes selenocysteine

			df = pandas.concat([aa2codons, aa2trna[organelle]], axis = 1).dropna(how = 'any').explode(1)

			# Check amino acids
			for aa in canonical_aas:
				if aa in aa2trna[organelle].index:
					pass
				else:
					logging.warning('At least one tRNA-{:s} gene is missing in the GenBank file. A \'generic_tRNA_triplet_aa\' dummy metabolite will be created to account for the related aminoacyl-tRNA synthetase expression.'.format(aa))

			trna_to_codon = { k:v + ['START'] if k in me_model.global_info['START_tRNA'] else v for k,v in zip(df[1].values, df[0].values) }

			# Read OSM to override data from the assumption "A tRNA-aa decodes all the codons for the specific amino acid"
			with open(me_model.global_info['df_gene_cplxs_mods_rxns'], 'rb') as infile:
				df_data = pandas.read_excel(infile)

			user_input = df_data[['Gene Locus ID', 'tRNA-codon association']].dropna(axis = 0, how = 'any').set_index('Gene Locus ID')
			user_input = user_input.to_dict()['tRNA-codon association']
			user_input = { k:v.split(',') for k,v in user_input.items() }

			if not bool(user_input): # an empty dictionary is False
				logging.warning('User input did not provide tRNA to codon associations and the derived from the GenBank file will be used.')
			else:
				# Here we replace inferred data with user input. User input should be a subset of the inferred data
				for trna, codons in trna_to_codon.items():
					if not set(user_input.get(trna, codons)) == set(codons):
						trna_to_codon[trna] = user_input.get(trna, codons)
						logging.warning('{} replaced with {}'.format(','.join(codons), ','.join(user_input[trna])))

			me_model.global_info['trna_to_codon'][organelle] = trna_to_codon

	# DO NOT REMOVE FOR-LOOP
	for organelle, trna_to_codon in me_model.global_info['trna_to_codon'].items():
		convert_aa_codes_and_add_charging(me_model, me_model.global_info['trna_to_aa'], trna_to_codon, organelle, verbose = verbose)

	if update:
		# Update all newly added reactions
		for r in tqdm.tqdm(list(me_model.reactions), 'Updating all TranslationReaction and TranscriptionReaction...', bar_format = bar_format):
			if isinstance(r, coralme.core.reaction.TranscriptionReaction):
				r.update(verbose = verbose)
			if isinstance(r, coralme.core.reaction.TranslationReaction):
				r.update(verbose = verbose)

def update_genbank_reactions(me_model, verbose = True):
	for r in me_model.reactions:
		if isinstance(r, (coralme.core.reaction.TranscriptionReaction, coralme.core.reaction.TranslationReaction)):
			r.update(verbose = verbose)

def add_m_model_content(me_model, m_model, complex_metabolite_ids = []):
	"""
	Add metabolite and reaction attributes to me_model from m_model. Also
	creates StoichiometricData objects for each reaction in m_model, and adds
	reactions directly to me_model if they are exchanges or demands.

	Parameters
	----------
	me_model : :class:`coralme.core.model.MEModel`
		The MEModel object to which the content will be added

	m_model : :class:`coralme.core.model.Model`
		The m_model which will act as the source of metabolic content for
		MEModel

	complex_metabolite_ids : list
		List of complexes which are 'metabolites' in the m-model reaction
		matrix, but should be treated as complexes.

	"""
	for met in tqdm.tqdm(m_model.metabolites, 'Adding Metabolites from M-model into the ME-model...', bar_format = bar_format):
		if met.id in complex_metabolite_ids:
			new_met = coralme.core.component.Complex(met.id)
		elif met.id.startswith('RNA_'):
			logging.warning('A metabolite with an invalid name (\'{:s}\') was added from the M-model or the \'reaction_matrix\' file. Please review it and correct it using the M-model or the \'me_metabolites\' file if necessary.'.format(met.id))
			new_met = coralme.core.component.Metabolite(met.id)
			##raise ValueError('Processed M-model should not contain RNAs ({:s})'.format(met.id))
			#new_met = me_model.metabolites.get_by_id(met.id)
		elif met.id.startswith('generic_tRNA'):
			new_met = coralme.core.component.GenerictRNA(met.id)
		elif met.id.startswith('generic_'):
			new_met = coralme.core.component.GenericComponent(met.id)
		else:
			new_met = coralme.core.component.Metabolite(met.id)

		new_met.name = met.name
		new_met.formula = met.formula
		new_met.compartment = met.compartment
		new_met.charge = met.charge
		new_met.annotation = met.annotation
		new_met.notes = met.notes
		me_model.add_metabolites(new_met)

	for reaction in tqdm.tqdm(m_model.reactions, 'Adding Reactions from M-model into the ME-model...', bar_format = bar_format):
		if reaction.id.startswith('BIOMASS_'):
			continue

		if reaction.id.startswith(('EX_', 'DM_', 'SK_', 'sink_')):
			new_reaction = coralme.core.reaction.MEReaction(reaction.id)
			me_model.add_reactions([new_reaction])
			new_reaction.subsystem = reaction.subsystem
			new_reaction.lower_bound = reaction.lower_bound
			new_reaction.upper_bound = reaction.upper_bound
			for met, stoichiometry in reaction.metabolites.items():
				new_reaction.add_metabolites({ me_model.metabolites.get_by_id(met.id): stoichiometry })
		else:
			reaction_data = coralme.core.processdata.StoichiometricData(reaction.id, me_model)
			reaction_data.lower_bound = reaction.lower_bound
			reaction_data.upper_bound = reaction.upper_bound
			reaction_data._stoichiometry = { k.id:v for k,v in reaction.metabolites.items() }

def add_dummy_reactions(me_model, transl_table, update = True):
	"""
	Add all reactions necessary to produce a dummy reaction catalyzed by
	'CPLX_dummy'.

	Parameters
	----------
	me_model : :class:`coralme.core.model.MEModel`
		The MEModel object to which the content will be added

	dna_seq : str
		DNA sequence of dummy gene. Should be representative of the average
		codon composition, amino acid composition, length of a gene in the
		organism being modeled

	update : bool
		If True, run update functions on all transcription, translation,
		complex formation, and metabolic reactions added when constructing
		dummy reactions.

	"""
	#stop_codons = me_model.global_info['stop_codons']
	stop_codons = Bio.Data.CodonTable.generic_by_id[list(transl_table)[0]].stop_codons
	stop_codons = set([ x.replace('U', 'T') for x in stop_codons ])

	df_codons = pandas.DataFrame(data = me_model.global_info['codon_usage'].values(), index = me_model.global_info['codon_usage'].keys())
	df_codons['per_1000'] = df_codons / df_codons.sum() * 1000

	# protein sequence based on codon usage
	seq = 'ATG'
	for codon, row in df_codons.iterrows():
	#     if row.amino_acid == 'Stop':
		if codon in stop_codons:
			continue
		seq += codon * int(row.per_1000 // 3)  # want roughly 300 aa
	# get the most used stop codon; old code
	# seq += df_codons[df_codons.amino_acid == 'Stop'].sort_values('per_1000').index[-1]
	seq += df_codons[df_codons.index.isin(stop_codons)].sort_values('per_1000').index[-1]

	dummy = coralme.core.processdata.StoichiometricData(me_model.global_info['dummy_rxn_id'], me_model)
	dummy.lower_bound = 0
	dummy.upper_bound = 1000
	dummy._stoichiometry = {'CPLX_dummy': -1}

	create_transcribed_gene(me_model, 'dummy', 'mRNA', seq)
	add_transcription_reaction(me_model, 'TU_RNA_dummy', {'dummy'}, seq)
	me_model.add_metabolites(coralme.core.component.TranslatedGene('protein_dummy'))

	transl_table = list(me_model.global_info['transl_tables']['c'])[0]
	prot_sequence = str(Bio.Seq.Seq(seq).translate(table = transl_table))
	add_translation_reaction(me_model, 'dummy', dna_sequence = seq, prot_sequence = prot_sequence, transl_table = transl_table, pseudo = True, update = update)

	try:
		complex_data = coralme.core.processdata.ComplexData('CPLX_dummy', me_model)
	except ValueError:
		logging.warning('Complex \'CPLX_dummy\' already present in the ME-model.')
		complex_data = me_model.process_data.get_by_id('CPLX_dummy')
	complex_data.stoichiometry = {'protein_dummy': 1}

	if update:
		complex_data.create_complex_formation()

def add_complex_to_model(me_model, complex_id, complex_stoichiometry, complex_modifications = None):
	"""
	Adds ComplexData to the model for a given complex.

	Parameters
	----------
	me_model : :class:`coralme.core.model.MEModel`

	complex_id : str
		ID of the complex and thus the model ComplexData

	complex_stoichiometry : dict
		{complex_id: {protein_<locus_tag>: stoichiometry}}

	complex_modifications : dict
		{subreaction_id: stoichiometry}

	"""

	if not complex_modifications:
		complex_modifications = {}

	complex_data = coralme.core.processdata.ComplexData(complex_id, me_model)
	# must add update stoichiometry one by one since it is a defaultdict
	for metabolite, value in complex_stoichiometry.items():
		complex_data.stoichiometry[metabolite] += value
	for modification, value in complex_modifications.items():
		complex_data.subreactions[modification] = value

def add_subreaction_data(me_model, modification_id, modification_stoichiometry, modification_enzyme = None, verbose = True):
	"""
	Creates a SubreactionData object for each modification defined by the
	function inputs.

	It's assumed every complex modification occurs spontaneously, unless a
	modification_enzyme argument is passed.

	If a modification uses an enzyme this can be updated after the
	SubreactionData object is already created

	Parameters
	----------
		me_model : :class:`coralme.core.model.MEModel`

	"""

	if modification_id in me_model.process_data:
		#if verbose:
		logging.warning('SubReaction \'{:s}\' is already in the ME-model and its stoichiometry was modified on your request.'.format(modification_id))
		me_model.process_data.get_by_id(modification_id).stoichiometry = modification_stoichiometry
		#else:
			#pass
	else:
		modification_data = coralme.core.processdata.SubreactionData(modification_id, me_model)
		modification_data.stoichiometry = modification_stoichiometry
		modification_data.enzyme = modification_enzyme
		try:
			modification_data._element_contribution = modification_data.calculate_element_contribution()
		except:
			modification_data._element_contribution = {}
		logging.warning('SubReaction \'{:s}\' was created in the ME-model.'.format(modification_id))

def add_model_complexes(me_model, complex_stoichiometry_dict, complex_modification_dict, verbose = True):
	"""
	Construct ComplexData for complexes into MEModel from its subunit
	stoichiometry, and a dictionary of its modification metabolites.

	It is assumed that each modification adds one equivalent of the
	modification metabolite. Multiple

	Intended to be used as a function for large-scale complex addition.

	For adding individual ComplexData objects, use add_complex_to_model

	Parameters
	----------
	me_model : :class:`coralme.core.model.MEModel`

	complex_stoichiometry_dict : dict
		{unmodified_complex_id: {protein_<locus_tag>: stoichiometry}}

	complex_modification_dict : dict
		{modified_complex_id:{
			core_enzyme: unmodified_complex_id,
			'modifications: {
				mod_metabolite: stoichiometry
				}
			}
		}

	"""
	for complex_id, stoichiometry in complex_stoichiometry_dict.items():
		add_complex_to_model(me_model, complex_id, stoichiometry, {})

	for modified_complex_id, info in complex_modification_dict.items():
		modification_dict = {}
		for metabolite, number in info['modifications'].items():
			if me_model.process_data.has_id('mod_' + metabolite):
				pass
			else:
				modification_id = 'mod_' + metabolite
				# add modification as a SubReaction
				add_subreaction_data(me_model, modification_id, {metabolite: -1.}, verbose = verbose)
				# stoichiometry of modification determined in modification_data.stoichiometry
				modification_dict[modification_id] = float(abs(number))

		core_enzyme = complex_modification_dict[modified_complex_id]['core_enzyme']
		stoichiometry = complex_stoichiometry_dict[core_enzyme]

		add_complex_to_model(me_model, modified_complex_id, stoichiometry, complex_modifications = modification_dict)

def add_metabolic_reaction_to_model(me_model, stoichiometric_data_id, directionality, complex_id = None, spontaneous = False, update = False, keff = 65.):
	"""
	Creates and add a MetabolicReaction to a MEModel.

	Parameters
	----------
	me_model : :class:`coralme.core.model.MEModel`
		MEModel that the MetabolicReaction will be added to

	stoichiometric_data_id : str
		ID of the StoichiometricData for the reaction being added

	directionality : str
		- Forward: Add reaction that occurs in the forward direction
		- Reverse: Add reaction that occurs in the reverse direction

	complex_id : str or None
		ID of the ComplexData for the enzyme that catalyze the reaction
		being added.

	spontaneous : bool
		- If True and complex_id='' add reaction as spontaneous reaction
		- If False and complex_id='' add reaction as orphan (CPLX_dummy
		  catalyzed)

	"""
	# Get stoichiometric data for reaction being added

	#print(stoichiometric_data_id)

	try:
		stoichiometric_data = me_model.process_data.get_by_id(stoichiometric_data_id)
	except KeyError:
		raise Exception('StoichiometricData for \'{:s}\' has not been added to ME-model.'.format(stoichiometric_data_id))

	# Get complex data and id based on arguments passed into function
	if isinstance(complex_id, str) and complex_id != 'dummy_MONOMER': # WARNING: Shouldn't it be "CPLX_dummy"?
		complex_data = me_model.process_data.get_by_id(complex_id)

	elif complex_id is None and spontaneous is True:
	#elif (complex_id == 'dummy_MONOMER' or complex_id is None) and spontaneous is True:
		complex_id = 'SPONT'
		complex_data = None
	elif complex_id is None and spontaneous is False:
	#elif (complex_id == 'dummy_MONOMER' or complex_id is None) and spontaneous is False:
		complex_id = 'CPLX_dummy'
		try:
			complex_data = me_model.process_data.get_by_id(complex_id)
		except KeyError:
			raise Exception('\'CPLX_dummy\' must be added to complex data to add orphan reactions.')
	else:
		raise ValueError('Complex id \'{:s}\' must be a string or None.'.format(str(complex_id)))

	if directionality.lower() == 'forward':
		direction = '_FWD_'
		reverse_flag = False
	elif directionality.lower() == 'reverse':
		direction = '_REV_'
		reverse_flag = True
	else:
		raise NameError('Reaction direction must be \'forward\' or \'reverse\'.')

	r = coralme.core.reaction.MetabolicReaction(''.join([stoichiometric_data_id, direction, complex_id]))
	me_model.add_reactions([r])
	#r.keff = keff
	r.coupling_coefficient_enzyme = keff # This will set mu/keff in per hour
	r.stoichiometric_data = stoichiometric_data
	r.reverse = reverse_flag

	if complex_data is not None:
		r.complex_data = complex_data
	if update:
		r.update(verbose = True)

def add_reactions_from_stoichiometric_data(
	me_model, rxn_to_cplx_dict,
	rxn_info_frame = pandas.DataFrame(), is_spontaneous = [],
	update = False, keff = 65.):
	"""
	Creates and adds MetabolicReaction for all StoichiometricData in model.

	Intended for use when adding all reactions from stoichiometric data for the
	first time.

	For adding an individual reaction use add_metabolic_reaction_to_model()

	Parameters
	----------
	me_model : :class:`coralme.core.model.MEModel`
		MEModel that the MetabolicReaction will be added to

	rxn_to_cplx_dict : dict
		{StoichiometricData.id: catalytic_enzyme_id}

	rxn_info_frame: :class:`pandas.Dataframe`
		Contains the ids, names and reversibility for each reaction in the
		metabolic reaction matrix as well as whether the reaction is
		spontaneous
	"""
	for reaction_data in tqdm.tqdm(list(me_model.stoichiometric_data), 'Processing StoichiometricData in ME-model...', bar_format = bar_format):
		#try:
			#spontaneous_flag = rxn_info_frame.is_spontaneous[reaction_data.id]
		#except KeyError:
			#spontaneous_flag = False
			#logging.warning('Reaction \'{:s}\' not in rxn_info_frame assumed nonspontaneous.'.format(reaction_data.id))

		#if spontaneous_flag == 1:
			#spontaneous = True
		#elif spontaneous_flag == 0:
			#spontaneous = False
		#else:
			#raise Exception('is_spontaneous must be \'True\' or \'False\'.')
		#TODO: detect 'spontaneous' from reaction name
		if reaction_data.id in is_spontaneous:
			spontaneous = True
		else:
			spontaneous = False

		# Reactions can be catalyzed by multiple isozymes so retrieve list of
		# complexes that catalyze the reaction
		complexes_list = rxn_to_cplx_dict.get(reaction_data.id, [None])

		# Add metabolic reactions for each isozyme
		for complex_id in complexes_list:

			directionality_list = []
			if reaction_data.lower_bound <= 0:
				directionality_list.append('reverse')
			if reaction_data.upper_bound >= 0:
				directionality_list.append('forward')
			if reaction_data.upper_bound == 0 and reaction_data.lower_bound == 0:
				#directionality_list.append('forward')
				logging.warning('Reaction \'{:s}\' cannot carry flux. Please check if it is the correct behavior.'.format(reaction_data.id))

			for directionality in directionality_list:
				add_metabolic_reaction_to_model(
					me_model,
					reaction_data.id,
					directionality,
					complex_id = complex_id,
					spontaneous = spontaneous,
					update = update,
					keff = keff
					)




import Bio
import copy
import pandas
import sympy

import cobra
import coralme

import logging
log = logging.getLogger(__name__)

import collections

class ProcessData(object):
	"""Generic class for storing information about a process

	This class essentially acts as a database that contains all of the
	relevant information needed to construct a particular reaction. For
	example, to construct a transcription reaction, following information must
	be accessed in some way:

	 - nucleotide sequence of the transcription unit
	 - RNA_polymerase (w/ sigma factor)
	 - RNAs transcribed from transcription unit
	 - other processes involved in transcription of RNAs (splicing, etc.)

	ME-model reactions are built from information in these objects.

	Parameters
	----------
	id : str
		Identifier of the ProcessData instance.

	model : :class:`coralme.core.model.MEModel`
		ME-model that the ProcessData is associated with

	"""

	def __init__(self, id, model):
		self.id = id
		self._model = model
		# parents need to be updated every time the process is updated
		# a parent must have an update method
		self._parent_reactions = set()
		model.process_data.append(self)

	# WARNING: MODIFIED FUNCTION FROM COBRAPY
	def copy(self) -> "ProcessData":
		"""Copy a ProcessData.
		"""
		# no references to model when copying
		model = self._model
		self._model = None
		# now we can copy
		new_processdata = copy.deepcopy(self)
		# restore the references
		self._model = model
		if not hasattr(self, 'id'):
			print(new_processdata)
		return new_processdata

	@property
	def model(self):
		"""
		Get the ME-model the process data is associated with

		Returns
		-------
		:class:`coralme.core.model.MEModel
			ME-model that uses this process data
		"""
		return self._model

	@property
	def parent_reactions(self):
		"""
		Get reactions that the ProcessData instance is used to construct.

		Returns
		-------
		set
			Parent reactions of ProcessData

		"""
		return {self._model.reactions.get_by_id(i) for i in self._parent_reactions}

	def update_parent_reactions(self):
		"""

		Executes the update() function for all reactions that the ProcessData
		instance is used to construct.

		"""
		reactions = self._model.reactions
		for i in self._parent_reactions:
			reactions.get_by_id(i).update()

	def __repr__(self):
		return '<{:s} {:s} at 0x{:x}>'.format(self.__class__.__name__, self.id, id(self))

class StoichiometricData(ProcessData):
	"""Encodes the stoichiometry for a metabolic reaction.

	StoichiometricData defines the metabolite stoichiometry and upper/lower
	bounds of metabolic reaction

	Parameters
	----------
	id : str
		Identifier of the metabolic reaction. Should be identical to the
		M-model reactions in most cases.

	model : :class:`coralme.core.model.MEModel`
		ME-model that the StoichiometricData is associated with

	Attributes
	----------
	_stoichiometry : dict
		Dictionary of {metabolite_id: stoichiometry} for reaction

	subreactions : :class:`collections.DefaultDict(int)`
		Cases where multiple enzymes (often carriers ie. Acyl Carrier Protein)
		are involved in a metabolic reactions.

	upper_bound : int
		Upper reaction bound of metabolic reaction. Should be identical to the
		M-model reactions in most cases.

	lower_bound : int
		Lower reaction bound of metabolic reaction. Should be identical to the
		M-model reactions in most cases.
	"""
	def __init__(self, id, model):
		ProcessData.__init__(self, id, model)
		self._stoichiometry = {}
		self.subreactions = collections.defaultdict(int)
		self.lower_bound = 0.
		self.upper_bound = 1000.

	@property
	def stoichiometry(self):
		"""
		Get or set metabolite stoichiometry for reaction.

		Returns
		-------
		dict
			Dictionary of {metabolite_id: stoichiometry}
		"""
		return self._stoichiometry

	@stoichiometry.setter
	def stoichiometry(self, value):
		if not isinstance(value, dict):
			raise TypeError('Stoichiometry must be a dict, not ({:s})'.format(type(value)))
		for k in value:
			if not isinstance(k, str):
				raise TypeError('Stoichiometry keys must be strings, not \'{:s}\'.'.format(type(k)))
		self._stoichiometry = value

class SubreactionData(ProcessData):
	"""
	Parameters
	----------
	id : str
		Identifier of the subreaction data. As a best practice, if the
		subreaction data details a modification, the ID should be prefixed
		with 'mod + _'

	model : :class:`coralme.core.model.MEModel`
		ME-model that the SubreactionData is associated with

	Attributes
	----------
	enzyme : list or str or None
		List of :attr:`coralme.core.component.Complex.id` s for enzymes that
		catalyze this process

		or

		String of single :attr:`coralme.core.component.Complex.id` for enzyme
		that catalyzes this process

	keff : float
		Effective turnover rate of enzyme(s) in subreaction process

	_element_contribution : dict
		If subreaction adds a chemical moiety to a macromolecules via a
		modification or other means, net element contribution of the
		modification process should be accounted for. This can be used to
		mass balance check each of the individual processes.

		Dictionary of {element: net_number_of_contributions}

	"""
	def __init__(self, id, model):
		ProcessData.__init__(self, id, model)
		self.stoichiometry = {}
		self.enzyme = None
		self._coupling_coefficient_subreaction = sympy.Mul(self._model.mu, sympy.Rational('1/3600'), model.symbols['k^default_cat']**-1, evaluate = False)
		self._element_contribution = {}

	# Backward compatibility
	@property
	def keff(self):
		"""
		returns the keff value, not the coupling coefficient, in per second
		"""
		value = self._model.mu * sympy.Rational('1/3600') / self._coupling_coefficient_subreaction
		try:
			return float(value)
		except:
			return float(value.xreplace(self._model.global_info['default_parameters']))

	# Backward compatibility
	@keff.setter
	def keff(self, value):
		"""
		value is the keff in per second, not the coupling coefficient in per hour
		this returns the coupling coefficient as growth rate divided by the keff, in per hour
		"""
		self._coupling_coefficient_subreaction = sympy.Mul(self._model.mu, sympy.Rational('1/3600'), value**-1, evaluate = False)

	@property
	def coupling_coefficient_subreaction(self):
		"""
		returns the coupling coefficient, not the keff value, in per hour
		"""
		return self._coupling_coefficient_subreaction

	@coupling_coefficient_subreaction.setter
	def coupling_coefficient_subreaction(self, value):
		"""
		value is the keff in per second, not the coupling coefficient in per hour
		this returns the coupling coefficient as growth rate divided by the keff, in per hour
		"""
		self._coupling_coefficient_subreaction = sympy.Mul(self._model.mu, sympy.Rational('1/3600'), value**-1, evaluate = False)

	@property
	def element_contribution(self):
		"""
		Get net contribution of elements from subreaction process to
		macromolecule

		If subreaction adds a chemical moiety to a macromolecules via a
		modification or other means, net element contribution of the
		modification process should be accounted for. This can be used to
		mass balance check each of the individual processes.

		Returns
		-------
		dict
			Dictionary of {element: net_number_of_contributions}

		"""
		if self._element_contribution:
			return self._element_contribution
		else:
			contribution = { k:v for k,v in self.calculate_element_contribution().items() if v }

		# Return 'trivial' cases (only one modifying metabolite in the
		# reactants and no products) without warning
		if len(self.stoichiometry) == 1 and list(self.stoichiometry.values())[0] < 0:
			return self.calculate_element_contribution()
		elif contribution:
			logging.warning('No element contribution input for SubReaction \'{:s}\'. Calculating based on stoichiometry instead.'.format(self.id))
			return self.calculate_element_contribution()
		else:
			return {}

	@element_contribution.setter
	def element_contribution(self, value):
		if not isinstance(value, dict):
			raise TypeError('Elemental_contribution must be a dictionary, not \'{:s}\'.'.format(type(value)))
		self._element_contribution = value

	def calculate_element_contribution(self):
		"""
		Calculate net contribution of chemical elements based on the
		stoichiometry of the subreaction data

		Returns
		-------
		dict
			Dictionary of {element: net_number_of_contributions}

		"""
		elements = collections.defaultdict(int)
		for met, coefficient in self.stoichiometry.items():
			if self._model.metabolites.has_id(met):
				met_obj = self._model.metabolites.get_by_id(met)
			else:
				logging.warning('The metabolite \'{:s}\' must exist in the ME-model to calculate the element contribution.'.format(met))
				continue

			# elements lost in conversion are added to complex, protein, etc.
			if not met_obj.elements and not isinstance(met_obj, coralme.core.component.GenerictRNA):
				logging.warning('Metabolite \'{:s}\' does not have a formula. If it is a \'Complex\', its formula will be determined from amino acid composition and prosthetic groups stoichiometry. Otherwise, please add it to the M-model.'.format(met_obj.id))

			for e, n in met_obj.elements.items():
				elements[e] -= n * coefficient

		return elements

	def calculate_biomass_contribution(self):
		"""
		Calculate net biomass increase/decrease as a result of the subreaction
		process.

		If subreaction adds a chemical moiety to a macromolecules via a
		modification or other means, the biomass contribution of the
		modification process should be accounted for and ultimately included
		in the reaction it is involved in.

		Returns
		-------
		float
			Mass of moiety transferred to macromolecule by subreaction

		"""
		elements = self.element_contribution

		# Create temporary metabolite for calculating formula weight
		tmp_met = cobra.Metabolite('mass')
		coralme.util.massbalance.elements_to_formula(tmp_met, elements)

		return tmp_met.formula_weight

	def get_complex_data(self):
		"""
		Get the complex data that the subreaction is involved in

		Yields
		------
		:class:`coralme.core.processdata.ComplexData`
			ComplexData that subreaction is involved in
		"""
		for i in self._model.complex_data:
			if self.id in i.subreactions:
				yield i

	def get_all_usages(self):
		"""
		Get all process data that the subreaction is involved in

		Yields
		------
		:class:`coralme.core.processdata.ProcessData`
			ProcessData that subreaction is involved in
		"""
		for i in self._model.process_data:
			if hasattr(i, 'subreactions') and self.id in i.subreactions:
				yield i

class ComplexData(ProcessData):
	"""Contains all information associated with the formation of an
	functional enzyme complex.

	This can include any enzyme complex modifications required for the enzyme
	to become active.

	Parameters
	----------
	id : str
		Identifier of the complex data. As a best practice, this should
		typically use the same ID as the complex being formed. In cases with
		multiple ways to form complex '_ + alt' or similar suffixes can be
		used.

	model : :class:`coralme.core.model.MEModel`
		ME-model that the ComplexData is associated with

	Attributes
	----------

	stoichiometry : :class:`collections.DefaultDict(int)`
		Dictionary containing {protein_id: count} for all protein subunits
		comprising enzyme complex

	subreactions : dict
		Dictionary of {subreaction_data_id: count} for all complex formation
		subreactions/modifications. This can include cofactor/prosthetic group
		binding or enzyme side group addition.

	"""

	def __init__(self, id, model):
		ProcessData.__init__(self, id, model)
		# {Component.id: stoichiometry}
		self.stoichiometry = collections.defaultdict(int)
		# {SubreactionData.id : number}
		# Forming some metacomplexes occur in multiple steps
		self.subreactions = {}
		self._complex_id = None  # assumed to be the same as id if None

	@property
	def formation(self):
		"""Get the formation reaction object

		Returns
		-------
		:class:`coralme.core.reaction.ComplexFormation`
			Complex formation reaction detailed in ComplexData
		"""
		try:
			return self._model.reactions.get_by_id('formation_' + self.id)
		except KeyError:
			return None

	@property
	def complex(self):
		"""
		Get complex metabolite object

		Returns
		-------
		:class:`coralme.core.component.Complex`
			Instance of complex metabolite that ComplexData is used to
			synthesize
		"""
		return self._model.metabolites.get_by_id(self.complex_id)

	@property
	def complex_id(self):
		"""
		Get  and set complex ID for product of complex formation reaction

		There are cases where multiple equivalent processes can result in
		the same final complex. This allows the equivalent final complex
		complex_id to be queried. This only needs set in the above case

		Returns
		-------
		str
			ID of complex that ComplexData is used to synthesize
		"""

		return self.id if self._complex_id is None else self._complex_id

	@complex_id.setter
	def complex_id(self, value):
		self._complex_id = None if value == self.id else value

	def create_complex_formation(self, verbose=True):
		"""creates a complex formation reaction

		This assumes none exists already. Will create a reaction (prefixed by
		'formation') which forms the complex

		Parameters
		----------
		verbose : bool
			If True, print if a metabolite is added to model during update

		"""
		formation_id = 'formation_' + self.id
		if formation_id in self._model.reactions:
			raise ValueError('Reaction \'{:s}\' already in the ME-model.'.format(formation_id))
		formation = coralme.core.reaction.ComplexFormation(formation_id)
		formation.complex_data_id = self.id
		formation._complex_id = self.complex_id
		self._model.add_reactions([formation])
		formation.update(verbose = verbose)

class TranscriptionData(ProcessData):
	"""
	Class for storing information needed to define a transcription reaction

	Parameters
	----------
	id : str
		Identifier of the transcription unit, typically beginning with 'TU'

	model : :class:`coralme.core.model.MEModel`
		ME-model that the TranscriptionData is associated with

	Attributes
	----------

	nucleotide_sequence : str
		String of base pair abbreviations for nucleotides contained in the
		transcription unit

	RNA_products : set
		IDs of :class:`coralme.core.component.TranscribedGene` that the
		transcription unit encodes. Each member should be prefixed with
		'RNA + _'

	RNA_polymerase : str
		ID of the :class:`coralme.core.component.RNAP` that transcribes the
		transcription unit. Different IDs are used for different sigma factors

	subreactions : :class:`collections.DefaultDict(int)`
		Dictionary of
		{:class:`coralme.core.processdata.SubreactionData` ID: num_usages}
		required for the transcription unit to be transcribed

	"""
	def __init__(self, id, model, nucleotide_sequence, rnap, rna_products, organelle):
		ProcessData.__init__(self, id, model)
		self.nucleotide_sequence = nucleotide_sequence
		self.RNA_products = rna_products
		self.original_RNA_products = rna_products
		self.RNA_polymerase = rnap
		self.organelle = organelle

		self._subreactions = collections.defaultdict(int)
		self._coupling_coefficient_rnapol = sympy.Mul(len(nucleotide_sequence), model.symbols['v_rnap'], evaluate = False)

	@property
	def coupling_coefficient_rnapol(self):
		return self._coupling_coefficient_rnapol

	@coupling_coefficient_rnapol.setter
	def coupling_coefficient_rnapol(self, value):
		self._coupling_coefficient_rnapol = value

	@property
	def n_cuts(self):
		# Number of cuts depends on the type of the RNAs in the TU
		return len([ x for x in self.RNA_types if x in ['rRNA', 'tRNA']]) * 2.

	@property
	def n_excised(self):
		# Number of excised bases depends on the type of the RNAs in the TU
		if set(self.RNA_types) == {'mRNA'}:
			return 0
		else:
			return sum(self.excised_bases.values())

	@property
	def n_overlapping(self):
		if self.id == 'RNA_dummy' or len(self.RNA_products) == 0:
			return 0

		import pyranges

		ranges = []
		for rna in self.RNA_products:
			data = self.model.metabolites.get_by_id(rna)
			left_pos = data.left_pos[0].replace('>', '').replace('<', '')
			right_pos = data.right_pos[0].replace('>', '').replace('<', '')
			ranges.append(['X', left_pos, right_pos, data.strand])

		df = pandas.DataFrame(ranges, columns = ['Chromosome', 'Start', 'End', 'Strand'])
		ranges = pyranges.PyRanges(df)

		# add overlapping ranges to df
		res = ranges.intersect(ranges, strandedness = 'same').df

		# remove original ranges
		# WARNING: What does happen if a gene overlaps completely another?
		tmp = pandas.merge(res, ranges.df, how = 'outer', indicator = True)
		tmp = tmp[tmp['_merge'] == 'left_only'].drop_duplicates()

		# return total length of the overlaps
		return abs(tmp['Start'] - tmp['End']).sum().sum()

	@property
	def subreactions(self):
		data = self._subreactions

		# Number of cuts and excised bases depend on the type of the RNAs in the TU
		if 'rRNA' not in set(self.RNA_types) and 'tRNA' not in set(self.RNA_types):
			return data

		n_overlapping = self.n_overlapping
		n_excised = self.n_excised
		n_cuts = self.n_cuts

		# WARNING: Because first 'if', n_cuts cannot be zero
		#if n_excised == 0 or (n_excised + n_overlapping) == 0 or n_cuts == 0:
			#return data
		#if n_excised == 0:
			#n_cuts = 0

		rna_types = list(self.RNA_types)
		n_trna = rna_types.count('tRNA')

		if 'rRNA' in set(rna_types):
			data['rRNA_containing_excision'] = n_cuts
		elif n_trna == 1:
			data['monocistronic_excision'] = n_cuts
		elif n_trna > 1:
			data['polycistronic_wout_rRNA_excision'] = n_cuts
		else: # only applies to rnpB (RNase P catalytic RNA component)
			data['monocistronic_excision'] = n_cuts

		# The non functional RNA segments need degraded back to nucleotides
		# TODO check if RNA_degradation requirement is per nucleotide
		data['RNA_degradation_machine'] = n_cuts
		data['RNA_degradation_atp_requirement'] = n_excised + n_overlapping

		return { k:v for k,v in data.items() if v != 0 }

	@property
	def nucleotide_count(self):
		"""
		Get count of each nucleotide contained in the nucleotide sequence

		Returns
		-------
		dict
			{nucleotide_id: number_of_occurences}

		"""
		#return { coralme.util.dogma.transcription_table[i]: self.nucleotide_sequence.count(i) for i in ['A', 'T', 'G', 'C'] }
		#return { coralme.util.dogma.transcription_table[k]:v for k,v in collections.Counter(self.nucleotide_sequence).items() }
		if self.organelle is None:
			if self._model.global_info['domain'].lower() in ['prokaryote', 'bacteria']:
				return { coralme.util.dogma.transcription_table['c'][k]:v for k,v in collections.Counter(self.nucleotide_sequence).items() }
			if self._model.global_info['domain'].lower() in ['eukarya', 'eukaryote']:
				return { coralme.util.dogma.transcription_table['n'][k]:v for k,v in collections.Counter(self.nucleotide_sequence).items() }
			#return { coralme.util.dogma.transcription_table['n'][k]:v for k,v in collections.Counter(self.nucleotide_sequence).items() }
		elif self.organelle.lower() in ['mitochondria', 'mitochondrion']:
			return { coralme.util.dogma.transcription_table['m'][k]:v for k,v in collections.Counter(self.nucleotide_sequence).items() }
		elif self.organelle.lower() in ['chloroplast', 'plastid']:
			return { coralme.util.dogma.transcription_table['h'][k]:v for k,v in collections.Counter(self.nucleotide_sequence).items() }
		else:
			logging.warning('The \'organelle\' property of the feature \'{:s}\' is not \'mitochondria\' or \'chloroplast\'.'.format(self.id))
			return { coralme.util.dogma.transcription_table['n'][k]:v for k,v in collections.Counter(self.nucleotide_sequence).items() }

	@property
	def RNA_types(self):
		"""
		Get generator consisting of the RNA type for each RNA product

		Yields
		------
		str
			(mRNA, tRNA, rRNA, ncRNA)
		"""
		for rna in self.RNA_products:
			rna_type = self._model.metabolites.get_by_id(rna).RNA_type
			if rna_type:
				yield rna_type

	@property
	def excised_bases(self):
		"""
		Get count of bases that are excised during transcription

		If a stable RNA (e.g. tRNA or rRNA) is coded for in the transcription
		unit, the transcript must be spliced in order for these to function.

		This determines whether the transcription unit requires splicing and,
		if so, returns the count of nucleotides within the transcription unit
		that are not accounted for in the RNA products, thus identifying the
		appropriate introns nucleotides.

		Returns
		-------
		dict
			{nucleotide_monophosphate_id: number_excised}

			i.e. {'amp_c': 10, 'gmp_c': 11, 'ump_c': 9, 'cmp_c': 11}

		"""
		rna_types = set(self.RNA_types)

		# Skip if TU does not have any annotated RNA Products
		if len(rna_types) == 0:
			return {'amp_c': 0, 'gmp_c': 0, 'ump_c': 0, 'cmp_c': 0}

		# Skip if TU only codes for mRNA
		# WARNING: The GenBank can contain other types of RNAs that break the condition of only mRNAs in the TU
		if rna_types == {'mRNA'}:
			return {'amp_c': 0, 'gmp_c': 0, 'ump_c': 0, 'cmp_c': 0}

		# WARNING: Features in the TU can overlap, thus this calculation must be corrected
		# Get dictionary of all nucleotide counts for TU
		seq = self.nucleotide_sequence
		#counts = { i: seq.count(i) for i in ('A', 'T', 'G', 'C') }
		counts = collections.Counter(seq)

		# Subtract bases contained in RNA_product from dictionary
		metabolites = self._model.metabolites
		for product_id in self.RNA_products:
			gene_seq = metabolites.get_by_id(product_id).nucleotide_sequence
			#for b in ('A', 'T', 'G', 'C'):
				#counts[b] -= gene_seq.count(b)
			counts.subtract(collections.Counter(gene_seq)) # inplace

		# First base being a triphosphate will be handled by the reaction
		# producing an extra ppi during transcription. But generally, we add
		# triphosphate bases when transcribing, but excise monophosphate bases.
		#monophosphate_counts = { coralme.util.dogma.transcription_table[k].replace('tp_c', 'mp_c'):v for k,v in counts.items() }
		monophosphate_counts = { coralme.util.dogma.transcription_table['c'][k].replace('tp_c', 'mp_c'):v for k,v in counts.items() }
		return monophosphate_counts

	@property
	def codes_stable_rna(self):
		"""
		Get whether transcription unit codes for a stable RNA

		Returns
		-------
		bool
			True if tRNA or rRNA in RNA products
			False if not

		"""
		has_stable_rna = False
		for RNA in self.RNA_products:
			try:
				gene = self._model.metabolites.get_by_id(RNA)
			except KeyError:
				pass
			else:
				if gene.RNA_type in ['tRNA', 'rRNA', 'ncRNA']:
					has_stable_rna = True
		return has_stable_rna

class GenericData(ProcessData):
	"""
	Class for storing information about generic metabolites

	Parameters
	----------
	id : str
		Identifier of the generic metabolite. As a best practice, this ID
		should be prefixed with 'generic + _'

	model : :class:`coralme.core.model.MEModel`
		ME-model that the GenericData is associated with

	component_list : list
		List of metabolite ids for all metabolites that can provide
		identical functionality
	"""
	def __init__(self, id, model, component_list):
		if not id.startswith('generic_'):
			logging.warning('Best practice for generic id to start with the \'generic_\' prefix.')
		ProcessData.__init__(self, id, model)
		self.component_list = component_list

		# bypass problems with GenericData not having complex and complex_id attributes
		self._complex_id = None  # assumed to be the same as id if None

	# bypass problems with GenericData not having complex and complex_id attributes
	@property
	def complex(self):
		"""
		Get complex metabolite object

		Returns
		-------
		:class:`coralme.core.component.Complex`
			Instance of complex metabolite that ComplexData is used to
			synthesize
		"""
		return self._model.metabolites.get_by_id(self.complex_id)

	@property
	def complex_id(self):
		"""
		Get  and set complex ID for product of complex formation reaction

		There are cases where multiple equivalent processes can result in
		the same final complex. This allows the equivalent final complex
		complex_id to be queried. This only needs set in the above case

		Returns
		-------
		str
			ID of complex that ComplexData is used to synthesize
		"""

		return self.id if self._complex_id is None else self._complex_id

	@complex_id.setter
	def complex_id(self, value):
		self._complex_id = None if value == self.id else value

	def create_reactions(self):
		"""

		Adds reaction with id '<metabolite_id> + _ + to + _ + <generic_id>'
		for each metabolite in self.component_list.

		Creates generic metabolite and generic reaction, if they do not already
		exist.
		"""
		model = self._model
		try:
			generic_metabolite = model.metabolites.get_by_id(self.id)
		except KeyError:
			generic_metabolite = coralme.core.component.GenericComponent(self.id)
			model.add_metabolites([generic_metabolite])
		for c_id in [ x for x in self.component_list if x.replace('RNA_', '') not in model.global_info['knockouts'] ]:
			reaction_id = c_id + '_to_' + self.id
			try:
				reaction = model.reactions.get_by_id(reaction_id)
			except KeyError:
				reaction = coralme.core.reaction.GenericFormationReaction(reaction_id)
				model.add_reactions([reaction])
			stoic = {
				generic_metabolite: 1,
				model.metabolites.get_by_id(c_id): -1
				}
			reaction.add_metabolites(stoic, combine=False)

class TranslationData(ProcessData):
	"""
	Class for storing information about a translation reaction.

	Parameters
	----------
	id : str
		Identifier of the gene being translated, typically the locus tag

	model : :class:`coralme.core.model.MEModel`
		ME-model that the TranslationData is associated with

	mrna : str
		ID of the mRNA that is being translated

	protein : str
		ID of the protein product.

	Attributes
	----------
	mRNA : str
		ID of the mRNA that is being translated

	protein : str
		ID of the protein product.

	subreactions : :class:`collections.DefaultDict(int)`
		Dictionary of
		{:attr:`coralme.core.processdata.SubreactionData.id`: num_usages}
		required for the mRNA to be translated

	nucleotide_sequence : str
		String of base pair abbreviations for nucleotides contained in the gene
		being translated

	"""
	def __init__(self, id, model, mrna, protein, nucleotide_sequence, organelle, translation, transl_table, pseudo):
		ProcessData.__init__(self, id, model)
		self.mRNA = mrna
		self.protein = protein
		self.nucleotide_sequence = nucleotide_sequence
		self.organelle = organelle
		self.translation = translation
		self.transl_table = transl_table
		self.pseudo = pseudo
		self.notes = []

		self.subreactions = collections.defaultdict(int)
		self._coupling_coefficient_ribosome = sympy.Mul(len(translation), model.symbols['v_ribo'], evaluate = False)
		self._coupling_coefficient_rna_synthesis = self._model.symbols['rna_amount'] + self._model.symbols['deg_amount']
		self._coupling_coefficient_hydrolysis = sympy.Mul((len(nucleotide_sequence) - 1) / 4., self._model.symbols['deg_amount'], evaluate = False) # deg_amount
		self._translational_efficiency = 1.

	@property
	def translational_efficiency(self):
		return self._translational_efficiency

	@translational_efficiency.setter
	def translational_efficiency(self, value):
		self._translational_efficiency = value

	@property
	def last_codon(self):
		"""
		Get the last codon contained in the mRNA sequence. This should
		correspond to the stop codon for the gene.

		Returns
		-------
		str
			Last 3 nucleotides comprising the last codon in the mRNA gene
			sequence

		"""
		return self.nucleotide_sequence[-3:].replace('T', 'U')

	@property
	def first_codon(self):
		"""
		Get the first codon contained in the mRNA sequence. This should
		correspond to the start codon for the gene.

		Returns
		-------
		str
			First 3 nucleotides comprising the first codon in the mRNA gene
			sequence

		"""
		return self.nucleotide_sequence[:+3].replace('T', 'U')

	def _itercodons(self):
		yield [i for i in self.codon_count]

	@property
	def sequence_as_codons(self):
		return [ self.nucleotide_sequence[i:i+3] for i in range(0, len(self.nucleotide_sequence), 3) ]

	@property
	def amino_acid_sequence(self):
		"""
		Get amino acid sequence from mRNA's nucleotide sequence

		Returns
		-------
		str
			Amino acid sequence

		"""
		#codons = (self.nucleotide_sequence[i: i + 3] for i in range(0, (len(self.nucleotide_sequence)), 3))
		#amino_acid_sequence = ''.join(coralme.util.dogma.codon_table[i] for i in codons)
		#amino_acid_sequence = str(Bio.Seq.Seq(self.nucleotide_sequence).translate(self._model.global_info['codon_table']))

		codons = self.sequence_as_codons # it includes the stop codon

		# translate first codon
		amino_acid_sequence = Bio.Seq.Seq('M')
		if isinstance(self.transl_table, str): # from JSON
			self.transl_table = Bio.Data.CodonTable.generic_by_id[int(list(self.transl_table)[0])]

		if codons[0] not in self.transl_table.start_codons:
			logging.warning('First codon in \'{:s}\' does not encode a start methionine. A methionine replaces the first amino acid.'.format(self.id))
			self.notes.append('Codon at position 0 does not encode a start methionine.')

		# translate rest of the sequence
		for idx, codon in enumerate(codons[1:-1]): # avoid start and stop codon
			if codon == 'TGA' and self.transl_table.id == 11:
				# Ser-tRNA is the precursor of Sec-tRNA.
				# Reaction Ser-tRNA(Sec) => Sec-tRNA(Sec) is added as a subreaction in translation reactions
				aa = 'S'
				logging.warning('Internal stop codon UGA identified in \'{:s}\' and translated into Selenocysteine tRNA precursor.'.format(self.id))
				self.notes.append('Codon at position {:d} encodes Selenocysteine.'.format(idx+1))
			elif codon in self._model.global_info['genetic_recoding'].keys():
				aa = '_' # placeholder to identify a recoded stop codon in self.amino_acid_sequence when creating TranslationReactions
				logging.warning('Internal stop codon \'{:s}\' identified in \'{:s}\' following user input.'.format(codon, self.id))
				self.notes.append('Codon at position {:d} encodes a recoded stop codon.'.format(idx+1))
			elif codon in self.transl_table.stop_codons:
				aa = '_' # placeholder to identify a recoded stop codon in self.amino_acid_sequence when creating TranslationReactions
				logging.warning('Internal stop codon \'{:s}\' identified in \'{:s}\'. Translation will not proceed. Please check if the gene is a pseudogene.'.format(codon, self.id))
				self.notes.append('Codon at position {:d} encodes an internal stop codon (\'{:s}\') not recoded.'.format(idx+1, codon))
				#break
			else:
				aa = Bio.Seq.Seq(codon).translate(self.transl_table)
			# append the translated nucleotide
			amino_acid_sequence += aa

		# last codon does not need translation unless is not a stop codon
		if codons[-1] not in self.transl_table.stop_codons:
			logging.warning('Last codon in \'{:s}\' does not encode a stop codon.'.format(self.id))
			self.notes.append('Codon at position {:d} does not encode a stop codon.'.format(idx+2))
			amino_acid_sequence += Bio.Seq.Seq(codons[-1]).translate(self.transl_table)

		self.notes = sorted(set(self.notes))

		#amino_acid_sequence = amino_acid_sequence.rstrip('*')

		if self.id != 'dummy':
			if amino_acid_sequence != self.translation.rstrip('*'):
				logging.warning('Protein sequence for \'{:s}\' from the GenBank file differs from the inferred from nucleotide sequence and translation table.'.format(self.id))

		# WARNING: This was replaced by code above
		#if '*' in amino_acid_sequence or 'U' in amino_acid_sequence: # translation of selenocysteine
			##amino_acid_sequence = amino_acid_sequence.replace('*', 'C') # Cysteine?
			#amino_acid_sequence = amino_acid_sequence.replace('*', 'S') # Ser-tRNA is the precursor of Sec-tRNA
			#amino_acid_sequence = amino_acid_sequence.replace('U', 'S') # Ser-tRNA is the precursor of Sec-tRNA

		return amino_acid_sequence

	@property
	def amino_acid_count(self):
		"""Get number of each amino acid in the translated protein

		Returns
		-------
		dict
			{amino_acid_id: number_of_occurrences}
		"""

		#aa_count = collections.defaultdict(int)
		#for i in self.amino_acid_sequence:
			#aa_count[coralme.util.dogma.amino_acids[i]] += 1
		#return aa_count

		# Set compartment
		if self.organelle is None:
			compartment = '_c'
		elif self.organelle.lower() in ['mitochondria', 'mitochondrion']:
			compartment = '_m'
		elif self.organelle.lower() in ['chloroplast', 'plastid']:
			compartment = '_h'

		#return { coralme.util.dogma.amino_acids[k] + compartment:v for k,v in collections.Counter(self.amino_acid_sequence).items() }

		precount = []
		# WARNING: sequence_as_codons includes the stop codon, therefor it is +1 longer that amino_acid_sequence
		for idx, (codon, amino_acid) in enumerate(zip(self.sequence_as_codons, self.amino_acid_sequence)):
			if amino_acid == '_' and codon == '' and self.transl_table.id == 11:
				precount.append('ser__L_c') # Ser in the precursor for Selenocysteine
			elif amino_acid == '_' and codon in self._model.global_info['genetic_recoding']:
				# TODO: set compartment of the recoded stop codon?
				precount.append(list(self._model.global_info['genetic_recoding'][codon].keys())[0])
			elif amino_acid == '_' and codon not in self._model.global_info['genetic_recoding']:
				logging.warning('Internal stop codon at position \'{:d}\' has no recoding alternative. Please set up \'genetic_recoding\' dictionary. '.format(idx))
			else:
				precount.append(coralme.util.dogma.amino_acids[amino_acid] + compartment)
		return collections.Counter(precount)

	@property
	def codon_count(self):
		"""
		Get the number of each codon contained within the gene sequence

		Returns
		-------
		dict
			{codon_sequence: number_of_occurrences}

		"""
		#codons = (self.nucleotide_sequence[i: i+3] for i in range(0, len(self.nucleotide_sequence), 3))
		#codon_count = collections.defaultdict(int)
		#for i in codons:
			#codon_count[i.replace('T', 'U')] += 1

		codons = self.sequence_as_codons
		codons = [ x.replace('T', 'U') for x in codons if len(x) == 3 ]
		codon_count = collections.Counter(codons)

		return codon_count

	@property
	def subreactions_from_sequence(self):
		"""
		Get subreactions associated with each tRNA/AA addition.

		tRNA activity is accounted for as subreactions. This returns the
		subreaction counts associated with each amino acid addition, based
		on the sequence of the mRNA.

		Returns
		-------
		dict
			{:attr:`coralme.core.processdata.SubreactionData.id`: num_usages}
		"""
		subreactions = {}

		#table = self._model.global_info['translation_table']
		# Trip first and last codon. Not translated during elongation
		codon_count = self.codon_count
		codon_count[self.first_codon] -= 1
		codon_count[self.last_codon] -= 1

		for codon, count in codon_count.items():
			if count == 0:
				continue

			codon = codon.replace('U', 'T')
			#if codon == 'TGA' and table == 11:

			#abbreviated_aa = coralme.util.dogma.codon_table[codon]
			#abbreviated_aa = Bio.Seq.Seq(codon).translate(self._model.global_info['codon_table'])
			abbreviated_aa = Bio.Seq.Seq(codon).translate(self.transl_table)
			# Filter out the compartment and stereochemistry from aa id
			aa = coralme.util.dogma.amino_acids.get(abbreviated_aa, 'STOP_CODON').split('_')[0]

			if aa == 'STOP':
				if str(abbreviated_aa) == '*' and codon in self._model.global_info.get('genetic_recoding', {}).keys():
					# perform recoding of internal stop codons
					aa = list(self._model.global_info['genetic_recoding'][codon].keys())[0].replace('__L_c', '')
					logging.warning('Recoded stop codon \'{:s}\' to \'{:s}__L_c\' in \'{:s}\'.'.format(codon, aa, self.id))
				elif str(abbreviated_aa) == '*' and codon == 'TGA' and self.transl_table.id == 11:
					logging.warning('Adding selenocysteine for \'{:s}\', following translation table {:d} (See more https://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi#SG{:d}).'.format(self.id, self.transl_table.id, self.transl_table.id))
					aa = 'sec'
				else:
					logging.warning('Internal stop codon \'{:s}\' detected in \'{:s}\'. Please review if the gene is a pseudogene.'.format(codon, self.id))

			if aa == 'STOP':
				break # do not remove break or STOP_addition_at_UAA or similar will be added

			codon = codon.replace('T', 'U')
			subreaction_id = aa + '_addition_at_' + codon
			#try:
				#self._model.process_data.get_by_id(subreaction_id)
			#except KeyError:
				#logging.warning('The tRNA SubReaction \'{:s}\' is not in the ME-model.'.format(subreaction_id))
			if self._model.process_data.has_id(subreaction_id):
				subreactions[subreaction_id] = count
			else:
				logging.warning('The tRNA SubReaction \'{:s}\' is not in the ME-model.'.format(subreaction_id))

		return subreactions

	def add_elongation_subreactions(self, elongation_subreactions=set()):
		"""
		Add all subreactions involved in translation elongation.

		This includes:

		 - tRNA activity subreactions returned with
		   :meth:`subreactions_from_sequence` which is called within this
		   function.

		 - Elongation subreactions passed into this function. These will be
		   added with a value of len(amino_acid_sequence) - 1 as these are
		   involved in each amino acid addition

		Some additional enzymatic processes are required for each amino acid
		addition during translation elongation

		Parameters
		----------
		elongation_subreactions : set
			Subreactions that are required for each amino acid addition

		"""

		for subreaction_id in elongation_subreactions:
			try:
				self._model.process_data.get_by_id(subreaction_id)
			except KeyError:
				logging.warning('Elongation SubReaction \'{:s}\' is not in ME-model. However, it can be added later.'.format(subreaction_id))
			else:
				# No elongation subreactions needed for start codon
				self.subreactions[subreaction_id] = len(self.amino_acid_sequence) - 1.

		for subreaction_id, value in self.subreactions_from_sequence.items():
			self.subreactions[subreaction_id] = value

	def add_initiation_subreactions(self, start_codons=set(), start_subreactions=set()):
		"""
		Add all subreactions involved in translation initiation.

		Parameters
		----------
		start_codons : set, optional
			Start codon sequences for the organism being modeled

		start_subreactions : set, optional
			Subreactions required to initiate translation, including the
			activity by the start tRNA

		"""
		#print(self.mRNA, type(self.mRNA), self.first_codon, type(self.first_codon))
		if self.first_codon not in start_codons:
			logging.warning('\'{:s}\' starts with \'{:s}\', which is not a start codon'.format(self.mRNA, str(self.first_codon)))

		for subreaction_id in start_subreactions:
			try:
				self._model.process_data.get_by_id(subreaction_id)
			except KeyError:
				logging.warning('Initiation SubReaction \'{:s}\' is not in the ME-model. However, it can be added later.'.format(subreaction_id))
			else:
				self.subreactions[subreaction_id] = 1.

	def add_termination_subreactions(self, translation_terminator_dict=None):
		"""
		Add all subreactions involved in translation termination.

		Parameters
		----------
		translation_terminator_dict : dict or None
			{stop_codon : enzyme_id_of_terminator_enzyme}

		"""
		if not translation_terminator_dict:
			translation_terminator_dict = {}
		last_codon = self.last_codon
		term_enzyme = translation_terminator_dict.get(last_codon, None)
		if term_enzyme:
			termination_subreaction_id = last_codon + '_' + term_enzyme + '_mediated_termination_c'
			try:
				self._model.process_data.get_by_id(termination_subreaction_id)
			except KeyError:
				logging.warning('Termination SubReaction \'{:s}\' is not in ME-model. However, it can be added later.'.format(termination_subreaction_id))
			else:
				self.subreactions[termination_subreaction_id] = 1.
		else:
			logging.warning('No termination enzyme for \'{:s}\'. Please review if the gene is a pseudogene.'.format(self.mRNA))

class tRNAData(ProcessData):
	"""
	Class for storing information about a tRNA charging reaction.

	Parameters
	----------
	id : str
		Identifier for tRNA charging process. As best practice, this should
		be follow 'tRNA + _ + <tRNA_locus> + _ + <codon>' template. If tRNA
		initiates translation, <codon> should be replaced with START.

	model : :class:`coralme.core.model.MEModel`
		ME-model that the tRNAData is associated with

	amino_acid : str
		Amino acid that the tRNA transfers to an peptide

	rna : str
		ID of the uncharged tRNA metabolite. As a best practice, this ID should
		be prefixed with 'RNA + _'

	Attributes
	----------
	subreactions : :class:`collections.DefaultDict(int)`
		Dictionary of
		{:attr:`coralme.core.processdata.SubreactionData.id`: num_usages}
		required for the tRNA to be charged

	synthetase : str
		ID of the tRNA synthetase required to charge the tRNA with an amino
		acid

	synthetase_keff : float
		Effective turnover rate of the tRNA synthetase

	"""

	def __init__(self, id, model, amino_acid, rna, codon):
		ProcessData.__init__(self, id, model)
		self.codon = codon
		self.amino_acid = amino_acid
		self.RNA = rna
		self.subreactions = collections.defaultdict(int)
		self.synthetase = None

		self._synthetase_keff = self._model.symbols['k^default_cat']

		self._coupling_coefficient_trna_keff = self._model.symbols['k_tRNA']
		self._coupling_coefficient_trna_amount = sympy.Mul(self._model.mu, self._coupling_coefficient_trna_keff**-1, evaluate = False)
		self._coupling_coefficient_synthetase = sympy.Mul(self._model.mu, sympy.Rational('1/3600'), self._synthetase_keff**-1, (1 + self._coupling_coefficient_trna_amount), evaluate = False)

		self.organelle = None

	@property
	def coupling_coefficient_trna_amount(self):
		return self._coupling_coefficient_trna_amount

	@coupling_coefficient_trna_amount.setter
	def coupling_coefficient_trna_amount(self, value):
		return NotImplemented # user should modify model.default_parameters

	# Backward compatibility
	@property
	def synthetase_keff(self):
		"""
		returns the synthetase keff value, not the coupling coefficient, in per second
		"""
		value = self._synthetase_keff
		try:
			return float(value)
		except:
			return float(value.xreplace(self._model.global_info['default_parameters']))

	# Backward compatibility
	@synthetase_keff.setter
	def synthetase_keff(self, value):
		"""
		value is the synthetase keff in per second, not the coupling coefficient in per hour
		this returns the coupling coefficient as growth rate divided by the keff times (1 + coupling_coefficient_trna_amount), in per hour
		"""
		self._coupling_coefficient_synthetase = sympy.Mul(self._model.mu, sympy.Rational('1/3600'), value**-1, (1 + self._coupling_coefficient_trna_amount), evaluate = False)

	@property
	def coupling_coefficient_synthetase(self):
		"""
		returns the coupling coefficient, not the synthetase keff value, in per hour
		"""
		return self._coupling_coefficient_synthetase

	@coupling_coefficient_synthetase.setter
	def coupling_coefficient_synthetase(self, value):
		"""
		value is the synthetase keff in per second, not the coupling coefficient in per hour
		this returns the coupling coefficient as growth rate divided by the keff times (1 + coupling_coefficient_trna_amount), in per hour
		"""
		self._coupling_coefficient_synthetase = sympy.Mul(self._model.mu, sympy.Rational('1/3600'), value**-1, (1 + self._coupling_coefficient_trna_amount), evaluate = False)

class TranslocationData(ProcessData):
	"""
	Class for storing information about a protein translocation pathway

	Parameters
	----------
	id : str
		Identifier for translocation pathway.

	model : :class:`coralme.core.model.MEModel`
		ME-model that the TranslocationData is associated with

	Attributes
	----------
	keff : float
		Effective turnover rate of the enzymes in the translocation pathway

	enzyme_dict : dict
		Dictionary containing enzyme specific information about the way it is
		coupled to protein translocation

		{enzyme_id: {length_dependent: <True or False>,
		 fixed_keff: <True or False>}}

	length_dependent_energy : bool
		True if the ATP cost of translocation is dependent on the length of
		the protein

	stoichiometry : dict
		Stoichiometry of translocation pathway, typically ATP/GTP hydrolysis

	"""

	def __init__(self, id, model):
		ProcessData.__init__(self, id, model)
		self.keff = 65.
		self.enzyme_dict = {}
		self.length_dependent_energy = False
		self.stoichiometry = {}

class PostTranslationData(ProcessData):
	"""
	Parameters
	----------
	id : str
		Identifier for post translation process.

	model : :class:`coralme.core.model.MEModel`
		ME-model that the PostTranslationData is associated with

	processed_protein : str
		ID of protein following post translational process

	preprocessed_protein : str
		ID of protein before post translational process

	Attributes
	----------
	translocation : set
		Translocation pathways involved in post translation reaction.

		Set of {:attr:`coralme.core.processdata.TranslocationData.id`}

	translocation_multipliers : dict
		Some proteins require different coupling of translocation enzymes.

		Dictionary of
		{:attr:`coralme.core.processdata.TranslocationData.id`: float}

	surface_area : dict
		If protein is translated into the inner or outer membrane, the surface
		area the protein occupies can be accounted for as well.

		Dictionary of {SA_+<inner_membrane or outer_membrane>: float}

	subreactions : :class:`collections.DefaultDict(int)`
		If a protein is modified following translation, this is accounted for
		here

		Dictionary of {subreaction_id: float}

	biomass_type : str
		If the subreactions add biomass to the translated gene, the
		biomass type (:attr:`coralme.core.compontent.Constraint.id`) of the
		modification must be defined.

	folding_mechanism : str
		ID of folding mechanism for post translation reaction

	aggregation_propensity : float
		Aggregation propensity for the protein

	keq_folding : dict
		Temperature dependant keq for folding protein

		Dictionary of {str(temperature): value}

	k_folding : dict
		Temperature dependant rate constant (k) for folding protein

		Dictionary of {str(temperature): value}

	propensity_scaling : float
		Some small peptides are more likely to be folded by certain
		chaperones. This is accounted for using propensity_scaling.

	"""

	def __init__(self, id, model, processed_protein, preprocessed_protein):
		ProcessData.__init__(self, id, model)
		self.processed_protein_id = processed_protein
		self.unprocessed_protein_id = preprocessed_protein

		# For translocation post translation reactions
		self.translocation = set()
		self.translocation_multipliers = {}
		self.surface_area = {}

		# For post translation modifications
		self.subreactions = collections.defaultdict(int)
		self.biomass_type = ''

		# For protein folding reactions (FoldME)
		self.folding_mechanism = ''
		self.aggregation_propensity = 0.
		self.keq_folding = {}
		self.k_folding = {}
		self.propensity_scaling = 1.
